import argparse
import os
import pathlib
import typing
import warnings
from typing import Dict, List, NamedTuple, Optional, Tuple, Type, Union

from typing_extensions import Annotated, get_args, get_origin

from chalk.features import Feature, Features, FeatureSetBase
from chalk.importer import get_directory_root, import_all_python_files_from_dir
from chalk.streams import Windowed
from chalk.utils.collection_type import GenericAlias

try:
    from types import UnionType
except ImportError:
    UnionType = None


class ParsedAnnotation(NamedTuple):
    annotation: str
    include_in_protocol_cls: bool


class ParsedFeaturesClass(NamedTuple):
    module: str
    annotations: Dict[str, ParsedAnnotation]


class StubGenerator:
    def __init__(self) -> None:
        self._parsed_feature_classes: Dict[str, ParsedFeaturesClass] = {}
        self._module_to_stubs: List[Type[Features]] = []
        self._imports: Dict[Tuple[str, ...], str] = {}  # Mapping of full module name to the name it is imported as

    def _add_import(
        self, item: Union[str, type, typing._SpecialForm, Windowed], import_as: Optional[str] = None
    ) -> str:
        """Add an import for item.

        If ``item`` is already imported, then it will not be imported again.
        Its existing import name will be returned

        Args:
            item: The thing to import. Should be the full path to the item (e.g. ``typing.Optional``)

        Returns:
            str: The name that ``item`` will be imported as.
        """
        if isinstance(item, typing._SpecialForm):
            item = str(item)
        if isinstance(item, Windowed):
            item = type(item)
        if isinstance(item, type):
            item = f"{item.__module__}.{item.__name__}"
        if not isinstance(item, str):
            raise TypeError(f"Invalid type for {item}: {type(item)}")
        item_split = tuple(item.split("."))
        if item_split[0] == "builtins":
            assert len(item_split) == 2
            return item_split[1]
        if item_split in self._imports:
            imported_name = self._imports[item_split]
            if import_as is not None and import_as != imported_name:
                raise ValueError(
                    f"Item {'.'.join(item_split)} already imported as {imported_name}; cannot be imported again as {import_as}"
                )
            return imported_name
        if import_as is None:
            import_as = self._get_import_name(item_split)
        self._imports[item_split] = import_as
        return self._imports[item_split]

    def _get_import_name(self, item_split: Tuple[str, ...]):
        if item_split[0] in ("typing", "typing_extensions"):
            assert len(item_split) == 2
            return item_split[1]
        item_split = tuple(x.replace("_", "__") for x in item_split)
        item = f"__stubgen_{'_'.join(item_split)}"
        return item

    def _register_imports_for_annotation(self, annotation: Union[type, Windowed, GenericAlias]) -> str:

        if annotation == type(None):
            return "None"
        origin, args = get_origin(annotation), get_args(annotation)
        if origin is None:
            assert args == ()
            # Likely a "simple" annotation, like an enum class
            if isinstance(annotation, Windowed):
                origin_name = self._add_import(annotation, import_as="Windowed")
                arg_name = self._register_imports_for_annotation(annotation._kind)
                return f"{origin_name}[{arg_name}]"
            elif annotation.__module__ == "builtins":
                assert isinstance(annotation, type)
                return annotation.__name__
            else:
                return self._add_import(annotation)
        # Some sort of generic class, like a collection, or optional
        if origin == UnionType:
            # Convert new-style type unions to old-style typing.Unions
            origin = typing.Union
        if origin == Annotated:
            # If annotated, only pay attention to the first arg -- the actual type
            return self._register_imports_for_annotation(args[0])
        origin_name = self._add_import(origin)
        arg_names = [self._register_imports_for_annotation(arg) for arg in args]
        return f"{origin_name}[" + ", ".join(arg_names) + "]"

    def register_features_class(self, features_cls: Type[Features]):
        annotations: Dict[str, ParsedAnnotation] = {}
        for f in features_cls.features:
            assert isinstance(f, Feature)
            assert f.typ is not None
            if f.typ.is_dataframe:
                self._add_import("chalk.features.DataFrame", "DataFrame")
                annotations[f.attribute_name] = ParsedAnnotation(
                    "DataFrame",
                    include_in_protocol_cls=True,
                )
            else:
                # Make a best-effort to figure out imports.
                annotations[f.attribute_name] = ParsedAnnotation(
                    self._register_imports_for_annotation(f.typ.parsed_annotation),
                    include_in_protocol_cls=not f.is_windowed_pseudofeature and not f.is_autogenerated,
                )
        if features_cls.__name__ in self._parsed_feature_classes:
            raise ValueError(
                f"Unable to generate stubs due to multiple features sets called {features_cls.__name__}. Feature set names must be unique. Please rename one of these feature sets."
            )
        self._parsed_feature_classes[features_cls.__name__] = ParsedFeaturesClass(
            module=features_cls.__module__,
            annotations=annotations,
        )

    def generate_features_decorator_stub_file(self):
        self._add_import("typing.Optional", "Optional")
        self._add_import("typing.Protocol", "Protocol")
        self._add_import("typing.Type", "Type")
        self._add_import("typing.overload", "overload")
        self._add_import("chalk.features.feature_set.FeaturesMeta", "FeaturesMeta")
        self._add_import("chalk.features.Features", "Features")
        self._add_import("chalk.utils.duration.Duration", "Duration")
        self._add_import("chalk.features.Tags", "Tags")
        lines = [
            "# AUTO-GENERATED FILE. Do not edit. Run chalkpy stubgen to generate.",
            "# fmt: off",
            "# isort: skip_file",
            "from __future__ import annotations",
            "",
        ]
        # sort the imports for consistency
        imports = sorted(self._imports.items())
        for imp, import_as in imports:
            module = imp[:-1]
            name = imp[-1]
            lines.append(f"from {'.'.join(module)} import {name} as {import_as}")
        lines.append("")

        # The overall strategy is to create 3 stub classes for each features class.
        #
        # In no particular order:
        #
        # Class 1: The metaclasss. The metaclass includes @property descriptors for each class variable.
        # These descriptors are used to annotation class attributes as Type[feature]
        # Pyright requires that all annotations are types, not instances.
        #
        # Class 2: The class. The class includes an __init__ signature that includes all features, with the
        # correct type annotation. All features have a default value of `...`, since Chalk does not require feature
        # sets to be complete. Inside the __init__ body, all instance attributes are annotated. Placing the annotation
        # here, as opposed to the class body, does not conflict with the metaclass annotations
        #
        # Class 3: The protocol class. Since the @features decorator is passed the end user's class definition, and returns the
        # Chalk-ified version of it, we cannot "import" the user's class definition. So, we define a protocol class that contains the
        # same instance attributes. The type checker can match the user's class against the protocol.
        #
        # An @overload is used to match each protocol class (class #3) to the typestub version of the class (class #2)
        # Since the first matching protocol will be used, the feature sets with the _most_ features should appear first
        # Also including the class name as part of the sort key for stability
        parsed_feature_classes = sorted(
            self._parsed_feature_classes.items(), key=lambda x: (len(x[1].annotations), x[0]), reverse=True
        )
        for feature_cls_name, features_cls in parsed_feature_classes:
            # First generate the metaclass definition
            lines.append(
                f"""\
class {feature_cls_name}Metaclass(FeaturesMeta):"""
            )
            for attribute, annotation in features_cls.annotations.items():
                lines.append(
                    f"""\
    @property
    def {attribute}(self) -> Type[{annotation.annotation}]: ...
"""
                )
            if len(features_cls.annotations) == 0:
                lines.append(
                    f"""\
    ...
"""
                )
            # Next, generate the class definition
            lines.append(
                f"""\
class {feature_cls_name}(Features, metaclass={feature_cls_name}Metaclass):"""
            )
            lines.append(
                f"""\
    def __init__(
        self,"""
            )
            for attribute, annotation in features_cls.annotations.items():
                lines.append(
                    f"""\
        {attribute}: {annotation.annotation} = ...,"""
                )
            lines.append(
                """\
    ):"""
            )
            for attribute, annotation in features_cls.annotations.items():
                lines.append(
                    f"""\
        self.{attribute}: {annotation.annotation}"""
                )
            if len(features_cls.annotations) == 0:
                lines.append(
                    f"""\
        ..."""
                )
            lines.append("")
            # Finally, generate the protocol class representing the features class that is being transformed
            lines.append(
                f"""\
class {feature_cls_name}Protocol(Protocol):"""
            )
            has_proto_annotation = False
            for attribute, annotation in features_cls.annotations.items():
                if not annotation.include_in_protocol_cls:
                    continue
                has_proto_annotation = True
                lines.append(
                    f"""\
    {attribute}: {annotation.annotation}"""
                )
            if not has_proto_annotation:
                lines.append(
                    f"""\
    ..."""
                )
            lines.append("")

        # Add overloads to the @features decorator to hint that it returns
        # instances of the classes defined above, not the classes defined in the user's module
        # This allows us to put all type stubs in one file, under the chalk/features/feature namespace,
        # rather than having to create a type stub for each user's feature class.
        for features_cls_name, features_cls in parsed_feature_classes:
            # For each features class, adding two overloads. The first overload transforms the definition
            # from the user's code into our stub. However, since we are technically importing the version that
            # has already been transformed with @features, we need to annotate how to handle a class that
            # has already been processed -- hence the second definition which is effectively a no-op.
            lines.append(
                f"""\
@overload
def features(item: Type[{features_cls_name}Protocol]) -> Type[{features_cls_name}]: ...
"""
            )

        # We also need to handle when the user annotates the features class with args, such as an owner or tags, e.g.
        # @features(owner=...)
        # class MyFeaturesClass:
        #     ...
        # To do this, we add one additional overload for features that returns a protocol class.
        # This protocol class implements __call__ with overloads, similar to the above
        lines.append(
            """\
@overload
def features(
    *,
    owner: Optional[str] = ...,
    tags: Optional[Tags] = ...,
    max_staleness: Optional[Duration] = ...,
    etl_offline_to_online: Optional[bool] = ...,
) -> __stubgen__features_proto: ...

class __stubgen__features_proto(Protocol):"""
        )
        for features_cls_name, features_cls in parsed_feature_classes:
            overload = (
                "@overload" if len(parsed_feature_classes) > 1 else ""
            )  # It is incorrect to use @overload if there is just one definition
            lines.append(
                f"""\
    {overload}
    def __call__(self, item: Type[{features_cls_name}Protocol]) -> Type[{features_cls_name}]: ...
"""
            )
        if len(parsed_feature_classes) == 0:
            lines.append(
                """\
    ...
"""
            )
        return lines


def configure_stubgen_argparse(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--root",
        type=str,
        default=None,
        help="Project root to scan for features definitions. By default, will find the directory root.",
    )
    parser.add_argument(
        "--stub_path",
        "-s",
        type=str,
        default=None,
        help="Folder containing custom type stubs. By default, will use the `typings` folder, relative to the project root",
    )


def run_stubgen(args: argparse.Namespace):
    root = args.root
    if root is None:
        root = get_directory_root() or pathlib.Path(os.getcwd())
    else:
        assert isinstance(root, str)
        root = os.path.abspath(root)
        root = pathlib.Path(root)
    stub_path = args.stub_path
    if stub_path is None:
        stub_path = str(root / "typings")
    assert isinstance(stub_path, str)
    failed_imports = import_all_python_files_from_dir(root)
    if len(failed_imports) > 0:
        warnings.warn(
            f"Stubs may be incomplete due to errors in loading the following files: {', '.join([x.filename for x in failed_imports])}"
        )
    stubgen = StubGenerator()
    for features_cls in FeatureSetBase.registry.values():
        stubgen.register_features_class(features_cls)
    lines = stubgen.generate_features_decorator_stub_file()
    folder = os.path.join(stub_path, "chalk", "features")
    os.makedirs(folder, exist_ok=True)
    output_file = os.path.join(folder, "feature_set_decorator.pyi")
    with open(output_file, "w+") as f:
        f.write("\n".join(lines))
    print(f"Successfully wrote type stubs to {output_file}")



import dataclasses
from typing import Optional

import pytest

from chalk import State, offline, online
from chalk.features import features, owner


@features
class Txn:
    id: str
    amount: float


def test_can_define_owner():
    @online(owner="andy@chalk.ai")
    def resolver_with_owner(id: Txn.id) -> Txn.amount:
        ...


def test_can_extract_resolver_owner():
    @online(owner="andy@chalk.ai")
    def resolver_with_owner_2(id: Txn.id) -> Txn.amount:
        ...

    assert owner(resolver_with_owner_2) == "andy@chalk.ai"



import tempfile
from typing import Iterable

import pytest

from chalk.sql import SQLiteFileSource, TableIngestProtocol

try:
    import dotenv
except ImportError:
    pass
else:
    dotenv.load_dotenv()


@pytest.fixture(scope="module")
def temp_sqlite_source() -> Iterable[TableIngestProtocol]:
    with tempfile.NamedTemporaryFile() as tmpfile:
        yield SQLiteFileSource(filename=tmpfile.name)



from chalk.features import Features, features, online


@features
class HomeFeatures:
    home_id: str
    address: str
    price: int
    sq_ft: int


@online
def get_address(hid: HomeFeatures.home_id) -> HomeFeatures.address:
    return "Bridge Street" if hid == 1 else "Filbert Street"


@online
def get_home_data(hid: HomeFeatures.home_id) -> Features[HomeFeatures.price, HomeFeatures.sq_ft]:
    return HomeFeatures(
        price=200_000,
        sq_ft=2_000,
    )


def test_single_output():
    assert get_address(2) == "Filbert Street"


def test_multiple_output():
    result = get_home_data(2)
    assert isinstance(result, HomeFeatures)
    assert result.price == 200_000
    assert result.sq_ft == 2_000

    assert result != HomeFeatures(
        address="hello",
        price=200_000,
        sq_ft=2_000,
    )

    assert result == HomeFeatures(
        price=200_000,
        sq_ft=2_000,
    )



from __future__ import annotations

# NOTE: The functions in this file are not tested in chalkpy
# The integration tests in engine/ validate the behavior of load_dataset
import base64
import json
from collections import OrderedDict
from concurrent.futures import Future, ThreadPoolExecutor
from enum import IntEnum
from typing import TYPE_CHECKING, Any, Dict, List, Mapping, Optional, Sequence, Union

from typing_extensions import assert_never

from chalk.features import Feature, FeatureWrapper, ensure_feature
from chalk.utils.collections import get_unique_item
from chalk.utils.missing_dependency import missing_dependency_exception

if TYPE_CHECKING:
    import polars as pl

from chalk.features.feature_set import FeatureSetBase
from chalk.features.pseudofeatures import CHALK_TS_FEATURE, ID_FEATURE, OBSERVED_AT_FEATURE, PSEUDONAMESPACE
from chalk.serialization.codec import FEATURE_CODEC

_DEFAULT_EXECUTOR = ThreadPoolExecutor(16)


class DatasetVersion(IntEnum):
    """Format of the parquet file. Used when loading a dataset so we know what format it is in"""

    # This is the format that bigquery dumps to when specifying an output bucket and output format
    # as part of an (async) query job
    # The output contains extra columns, and all column names are b32 encoded, because
    # bigquery does not support '.' in column names.
    # The client will have to decode column names before loading this data
    # All data, except for feature times, are json encoded
    BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES = 1

    # This is the format returned by the dataset writer in engine/
    DATASET_WRITER = 2

    # This format uses separate columns for the observed at and timestamp columns
    # The observed at column is the actual timestamp from when the observation was observed,
    # whereas the timestamp column is the original timestamp that the user requested
    BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2 = 3

    COMPUTE_RESOLVER_OUTPUT_V1 = 4


def _parallel_download(uris: List[str], executor: ThreadPoolExecutor) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[polars]")

    df_futures: list[Future[pl.DataFrame]] = []
    for uri in uris:
        df_futures.append(executor.submit(pl.read_parquet, uri))

    dfs = [df.result() for df in df_futures]
    dfs = [x.select(sorted(x.columns)) for x in dfs]
    df = pl.concat(dfs)
    return df


def _load_dataset_from_chalk_writer(uris: List[str]) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[polars]")
    # V1 datasets should contain just a single URI
    # This URI can be read directly
    # We need to filter the features to remove any pseudofeatures
    if len(uris) != 1:
        raise ValueError("v1 datasets should have just a single URI")
    df = pl.read_parquet(uris[0])
    return df


def _decode_col_name(col_name: str) -> str:
    x_split = col_name.split("_")
    if x_split[0] == "ca":
        assert len(x_split) == 2
        return x_split[1]
    elif x_split[0] == "cb":
        root_fqn_b32 = x_split[1]
        return base64.b32decode(root_fqn_b32.replace("0", "=").upper()).decode("utf8") + "_".join(x_split[2:])
    else:
        raise ValueError(f"Unexpected identifier: {x_split[0]}")


def _decode_column_names(column_names: List[str]) -> Mapping[str, str]:
    ans: Dict[str, str] = {}
    for x in column_names:
        if x.startswith("__"):
            if x in ("__id__", ID_FEATURE.fqn):
                ans[x] = ID_FEATURE.fqn
            elif x in ("__ts__", CHALK_TS_FEATURE.fqn):
                # Preserve these columns as-is to help with loading the timestamp
                ans[x] = CHALK_TS_FEATURE.fqn
            elif x in ("__observed_at__", "__oat__", OBSERVED_AT_FEATURE.fqn):
                # Preserve these columns as-is to help with loading the timestamp
                ans[x] = OBSERVED_AT_FEATURE.fqn
            # Drop all the other metadata columns
            continue
        feature_name = _decode_col_name(x)
        if any(feature_name.endswith(f".__{x}__") for x in ("oat", "rat", "observed_at", "replaced_observed_at")):
            # Drop the timestamp metadata from individual features
            continue
        ans[x] = feature_name
    return ans


def _json_decode(x: Optional[str]):
    if x is None:
        return None
    return json.loads(x)


def _load_dataset_bigquery(
    uris: List[str],
    executor: Optional[ThreadPoolExecutor],
    output_feature_fqns: Optional[Sequence[str]],
    output_ts: bool,
    output_id: bool,
    version: DatasetVersion,
):
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[polars]")
    # V2 datasets are in multiple files, and have column names encoded
    # due to DB limitations (e.g. bigquery does not support '.' in column names)
    # In addition, the datasets may contain extra columns (e.g. replaced observed at)
    # All values are JSON encoded
    if executor is None:
        executor = _DEFAULT_EXECUTOR
    df = _parallel_download(uris, executor)

    if version in (
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2,
    ):
        decoded_col_names = _decode_column_names(df.columns)

        # Select only the columns in decoded_col_names
        df = df.select(list(decoded_col_names.keys()))
        df = df.rename(dict(decoded_col_names))

        # Using an OrderedDict so the order will match the order the user set in the
        # output argument
        expected_cols: Dict[str, pl.Expr] = OrderedDict()
        id_col = pl.col(str(ID_FEATURE))
        if output_id:
            # All dataframes have an __id__ column
            expected_cols[str(ID_FEATURE)] = id_col.alias(str(ID_FEATURE))

        ts_col = pl.col(str(OBSERVED_AT_FEATURE)).fill_null(pl.col(str(CHALK_TS_FEATURE))).dt.with_time_zone("UTC")
        if output_ts:
            # For the ts feature, we want `__oat__` if not null else `__ts__`
            # it is not null; otherwise, we want
            expected_cols[str(CHALK_TS_FEATURE)] = ts_col.alias(str(CHALK_TS_FEATURE))

        if output_feature_fqns is None:
            # If not provided, return all columns, except for the OBSERVED_AT_FEATURE
            # (the REPLACED_OBSERVED_AT was already dropped in _decode_col_names)
            for x in df.columns:
                if x not in expected_cols and not x.startswith(f"{PSEUDONAMESPACE}."):
                    expected_cols[x] = pl.col(x)

        else:
            # Make a best-effort attempt to determine the pkey and ts column fqn from the  root namespace
            # of the other features
            root_ns = get_unique_item(
                [x.split(".")[0] for x in df.columns if not x.startswith(f"{PSEUDONAMESPACE}.")], "root_ns"
            )
            ts_feature = None
            pkey_feature = None
            if root_ns in FeatureSetBase.registry:
                features_cls = FeatureSetBase.registry[root_ns]
                ts_feature = features_cls.__chalk_ts__
                pkey_feature = features_cls.__chalk_primary__

            for x in output_feature_fqns:
                if x in expected_cols:
                    continue
                if x in df.columns:
                    if x == str(CHALK_TS_FEATURE):
                        expected_cols[x] = ts_col.alias(x)
                    else:
                        expected_cols[x] = pl.col(x)
                    continue
                if x == str(CHALK_TS_FEATURE) or (ts_feature is not None and x == str(ts_feature)):
                    # The ts feature wasn't returned as the ts feature, but we are able to figure it out from the graph
                    # Alias the ts_col as the ts fqn (or CHALK_TS_FEATURE fqn if that's what was passed in)
                    expected_cols[x] = ts_col.alias(x)
                    continue
                if pkey_feature is not None and x == str(pkey_feature):
                    expected_cols[x] = id_col.alias(x)
                    continue
                else:
                    # We should _never_ hit this as the query should have failed before results are returned
                    # if an invalid feature was requested
                    raise ValueError(f"Feature '{x}' was not found in the results.")

        df = df.select(list(expected_cols.values()))

    elif version == DatasetVersion.COMPUTE_RESOLVER_OUTPUT_V1:
        unique_features = set(df.select(pl.col("feature_name").unique())["feature_name"].to_list())
        cols = [pl.col("value").filter(pl.col("feature_name") == fqn).first().alias(fqn) for fqn in unique_features]

        df = df.groupby("pkey").agg(cols)
        decoded_stmts = []
        for col in df.columns:
            if col == "pkey":
                continue
            else:
                decoded_stmts.append(
                    pl.col(col).apply(_json_decode, return_dtype=FEATURE_CODEC.get_polars_dtype(fqn=col))
                )
        df = df.select(decoded_stmts)
        # it might be a good idea to remember that we used to rename this __id__ column to the primary key
        # We also need to remove columns like feature.__oat__ and feature.__rat__
        df = df.select([col for col in df.columns if not col.endswith("__")])
        return df.select(sorted(df.columns))

    else:
        raise ValueError(f"Unsupported version: {version}")

    # The parquet file is all JSON-encoded except for the ts column. That is, the only datetime column is for the timestamp,
    # and all other columns are strings
    decoded_stmts = []
    for col, dtype in zip(df.columns, df.dtypes):
        if isinstance(dtype, pl.Datetime):
            # Assuming that the only datetime column is for timestamps
            decoded_stmts.append(pl.col(col).dt.with_time_zone("UTC"))
        else:
            decoded_stmts.append(pl.col(col).apply(_json_decode))

    return df.select(decoded_stmts)


def load_dataset(
    uris: List[str],
    version: DatasetVersion | int,
    output_features: Optional[Sequence[Union[str, Feature, FeatureWrapper, Any]]] = None,
    output_id: bool = True,
    output_ts: bool = True,
    executor: Optional[ThreadPoolExecutor] = None,
) -> pl.DataFrame:
    try:
        import polars as pl
    except ImportError:
        raise missing_dependency_exception("chalkpy[polars]")
    del pl  # Unused
    if not isinstance(version, DatasetVersion):
        try:
            version = DatasetVersion(version)
        except ValueError:
            raise ValueError(
                f"The dataset version ({version}) is not supported by this installed version of the Chalk client. "
                "Please upgrade your chalk client and try again."
            )

    if version == DatasetVersion.DATASET_WRITER:
        return _load_dataset_from_chalk_writer(uris)
    output_feature_fqns = (
        None
        if output_features is None
        else [x if isinstance(x, str) else ensure_feature(x).root_fqn for x in output_features]
    )
    if version in (
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES,
        DatasetVersion.BIGQUERY_JOB_WITH_B32_ENCODED_COLNAMES_V2,
        DatasetVersion.COMPUTE_RESOLVER_OUTPUT_V1,
    ):
        return _load_dataset_bigquery(
            uris,
            executor,
            version=version,
            output_feature_fqns=output_feature_fqns,
            output_id=output_id,
            output_ts=output_ts,
        )
    assert_never(version)



from mypy import api


def test_mypy_plugin():
    result = api.run(["./tests/mypy/mypy_test.py", "--show-traceback"])

    if result[0]:
        print("\nType checking report:\n")
        print(result[0])  # stdout

    if result[1]:
        print("\nError report:\n")
        print(result[1])  # stderr

    print("\nExit status:", result[2])



from os import PathLike
from typing import Optional, Union, overload

from chalk.sql._internal.incremental import IncrementalSettings
from chalk.sql._internal.integrations.bigquery import BigQuerySourceImpl
from chalk.sql._internal.integrations.cloudsql import CloudSQLSourceImpl
from chalk.sql._internal.integrations.mysql import MySQLSourceImpl
from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl
from chalk.sql._internal.integrations.redshift import RedshiftSourceImpl
from chalk.sql._internal.integrations.snowflake import SnowflakeSourceImpl
from chalk.sql._internal.integrations.sqlite import SQLiteFileSourceImpl, SQLiteInMemorySourceImpl
from chalk.sql.finalized_query import FinalizedChalkQuery
from chalk.sql.protocols import (
    BaseSQLSourceProtocol,
    ChalkQueryProtocol,
    SQLSourceWithTableIngestProtocol,
    StringChalkQueryProtocol,
    TableIngestProtocol,
)


@overload
def SnowflakeSource(*, name: str) -> BaseSQLSourceProtocol:
    ...


@overload
def SnowflakeSource(
    *,
    account_identifier: str = ...,
    warehouse: str = ...,
    user: str = ...,
    password: str = ...,
    db: str = ...,
    schema: str = ...,
    role: str = ...,
) -> BaseSQLSourceProtocol:
    ...


def SnowflakeSource(
    *,
    name: Optional[str] = None,
    account_identifier: Optional[str] = None,
    warehouse: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    db: Optional[str] = None,
    schema: Optional[str] = None,
    role: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return SnowflakeSourceImpl(
        name=name,
        account_identifier=account_identifier,
        warehouse=warehouse,
        user=user,
        password=password,
        db=db,
        schema=schema,
        role=role,
    )


@overload
def PostgreSQLSource(
    *,
    host: str = ...,
    port: Union[int, str] = ...,
    db: str = ...,
    user: str = ...,
    password: str = ...,
) -> SQLSourceWithTableIngestProtocol:
    ...


@overload
def PostgreSQLSource(*, name: str) -> SQLSourceWithTableIngestProtocol:
    ...


def PostgreSQLSource(
    *,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    name: Optional[str] = None,
) -> TableIngestProtocol:
    return PostgreSQLSourceImpl(host, port, db, user, password, name)


@overload
def MySQLSource(
    *,
    host: str = ...,
    port: Union[int, str] = ...,
    db: str = ...,
    user: str = ...,
    password: str = ...,
) -> SQLSourceWithTableIngestProtocol:
    ...


@overload
def MySQLSource(*, name: str) -> SQLSourceWithTableIngestProtocol:
    ...


def MySQLSource(
    *,
    host: Optional[str] = None,
    port: Optional[Union[int, str]] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    name: Optional[str] = None,
) -> SQLSourceWithTableIngestProtocol:
    return MySQLSourceImpl(host, port, db, user, password, name)


def SQLiteInMemorySource(name: Optional[str] = None) -> SQLSourceWithTableIngestProtocol:
    return SQLiteInMemorySourceImpl(name)


def SQLiteFileSource(filename: Union[str, PathLike], name: Optional[str] = None) -> SQLSourceWithTableIngestProtocol:
    return SQLiteFileSourceImpl(filename, name)


@overload
def RedshiftSource(
    *,
    host: str = ...,
    db: str = ...,
    user: str = ...,
    password: str = ...,
) -> BaseSQLSourceProtocol:
    ...


@overload
def RedshiftSource(*, name: str) -> BaseSQLSourceProtocol:
    ...


@overload
def RedshiftSource() -> BaseSQLSourceProtocol:
    ...


def RedshiftSource(
    *,
    host: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
    name: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return RedshiftSourceImpl(host, db, user, password, name)


@overload
def BigQuerySource() -> BaseSQLSourceProtocol:
    ...


@overload
def BigQuerySource(*, name: str) -> BaseSQLSourceProtocol:
    ...


def BigQuerySource(
    *,
    name: Optional[str] = None,
    project: Optional[str] = None,
    dataset: Optional[str] = None,
    location: Optional[str] = None,
    credentials_base64: Optional[str] = None,
    credentials_path: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return BigQuerySourceImpl(
        name=name,
        project=project,
        dataset=dataset,
        location=location,
        credentials_base64=credentials_base64,
        credentials_path=credentials_path,
    )


@overload
def CloudSQLSource() -> BaseSQLSourceProtocol:
    ...


@overload
def CloudSQLSource(*, name: str) -> BaseSQLSourceProtocol:
    ...


def CloudSQLSource(
    *,
    name: Optional[str] = None,
    instance_name: Optional[str] = None,
    db: Optional[str] = None,
    user: Optional[str] = None,
    password: Optional[str] = None,
) -> BaseSQLSourceProtocol:
    return CloudSQLSourceImpl(
        name=name,
        instance_name=instance_name,
        db=db,
        user=user,
        password=password,
    )


__all__ = [
    "BaseSQLSourceProtocol",
    "BigQuerySource",
    "CloudSQLSource",
    "MySQLSource",
    "PostgreSQLSource",
    "RedshiftSource",
    "SQLiteFileSource",
    "SQLiteInMemorySource",
    "SnowflakeSource",
    "FinalizedChalkQuery",
    "ChalkQueryProtocol",
    "StringChalkQueryProtocol",
    "TableIngestProtocol",
    "SQLSourceWithTableIngestProtocol",
    "IncrementalSettings",
]



from unittest.mock import mock_open, patch

from chalk.gitignore.gitignore_parser import parse_gitignore


def test_simple():
    matches = _parse_gitignore_string(("__pycache__/\n" "*.py[cod]"), fake_base_dir="/home/michael")
    assert not matches("/home/michael/main.py")
    assert matches("/home/michael/main.pyc")
    assert matches("/home/michael/dir/main.pyc")
    assert matches("/home/michael/__pycache__")


def test_wildcard():
    matches = _parse_gitignore_string("hello.*", fake_base_dir="/home/michael")
    assert matches("/home/michael/hello.txt")
    assert matches("/home/michael/hello.foobar/")
    assert matches("/home/michael/dir/hello.txt")
    assert matches("/home/michael/hello.")
    assert not matches("/home/michael/hello")
    assert not matches("/home/michael/helloX")


def test_anchored_wildcard():
    matches = _parse_gitignore_string("/hello.*", fake_base_dir="/home/michael")
    assert matches("/home/michael/hello.txt")
    assert matches("/home/michael/hello.c")
    assert not matches("/home/michael/a/hello.java")


def test_trailingspaces():
    matches = _parse_gitignore_string(
        (
            "ignoretrailingspace \n"
            "notignoredspace\\ \n"
            "partiallyignoredspace\\  \n"
            "partiallyignoredspace2 \\  \n"
            "notignoredmultiplespace\\ \\ \\ "
        ),
        fake_base_dir="/home/michael",
    )
    assert matches("/home/michael/ignoretrailingspace")
    assert not matches("/home/michael/ignoretrailingspace ")
    assert matches("/home/michael/partiallyignoredspace ")
    assert not matches("/home/michael/partiallyignoredspace  ")
    assert not matches("/home/michael/partiallyignoredspace")
    assert matches("/home/michael/partiallyignoredspace2  ")
    assert not matches("/home/michael/partiallyignoredspace2   ")
    assert not matches("/home/michael/partiallyignoredspace2 ")
    assert not matches("/home/michael/partiallyignoredspace2")
    assert matches("/home/michael/notignoredspace ")
    assert not matches("/home/michael/notignoredspace")
    assert matches("/home/michael/notignoredmultiplespace   ")
    assert not matches("/home/michael/notignoredmultiplespace")


def test_comment():
    matches = _parse_gitignore_string(
        "somematch\n#realcomment\nothermatch\n\\#imnocomment", fake_base_dir="/home/michael"
    )
    assert matches("/home/michael/somematch")
    assert not matches("/home/michael/#realcomment")
    assert matches("/home/michael/othermatch")
    assert matches("/home/michael/#imnocomment")


def test_ignore_directory():
    matches = _parse_gitignore_string(".venv/", fake_base_dir="/home/michael")
    assert matches("/home/michael/.venv")
    assert matches("/home/michael/.venv/folder")
    assert matches("/home/michael/.venv/file.txt")


def test_ignore_directory_asterisk():
    matches = _parse_gitignore_string(".venv/*", fake_base_dir="/home/michael")
    assert not matches("/home/michael/.venv")
    assert matches("/home/michael/.venv/folder")
    assert matches("/home/michael/.venv/file.txt")


def test_negation():
    matches = _parse_gitignore_string(
        """
*.ignore
!keep.ignore
            """,
        fake_base_dir="/home/michael",
    )
    assert matches("/home/michael/trash.ignore")
    assert not matches("/home/michael/keep.ignore")
    assert matches("/home/michael/waste.ignore")


def test_double_asterisks():
    matches = _parse_gitignore_string("foo/**/Bar", fake_base_dir="/home/michael")
    assert matches("/home/michael/foo/hello/Bar")
    assert matches("/home/michael/foo/world/Bar")
    assert matches("/home/michael/foo/Bar")


def test_single_asterisk():
    matches = _parse_gitignore_string("*", fake_base_dir="/home/michael")
    assert matches("/home/michael/file.txt")
    assert matches("/home/michael/directory")
    assert matches("/home/michael/directory-trailing/")


def _parse_gitignore_string(data: str, fake_base_dir: str):
    with patch("builtins.open", mock_open(read_data=data)):
        success = parse_gitignore(f"{fake_base_dir}/.gitignore", fake_base_dir)
        return success



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any, List, Mapping, Optional, Protocol, Sequence, Union

import pandas as pd

from chalk.client.models import (
    ChalkError,
    FeatureResult,
    OfflineQueryContext,
    OnlineQueryContext,
    ResolverRunResponse,
    WhoAmIResponse,
)
from chalk.features import DataFrame, Feature, FeatureWrapper

if TYPE_CHECKING:
    import polars as pl


class OnlineQueryResponseProtocol(Protocol):
    # The output features and any query metadata
    data: List[FeatureResult]

    # Errors encountered while running the resolvers.
    # If no errors were encountered, this field is empty.
    errors: Optional[List[ChalkError]]

    def get_feature(self, feature: Any) -> Optional[FeatureResult]:
        """
        A convenience method for accessing feature result from the data response

        :param feature: The feature or its string representation
        :return: The FeatureResult for the feature, if it exists
        """
        ...

    def get_feature_value(self, feature: Any) -> Optional[Any]:
        """
        A convenience method for accessing feature values from the data response

        :param feature: The feature or its string representation
        :return: The value of the feature
        """
        ...


class ChalkAPIClientProtocol(Protocol):
    def trigger_resolver_run(
        self,
        resolver_fqn: str,
        context: Optional[OfflineQueryContext] = None,
    ) -> ResolverRunResponse:
        """
        Triggers a resolver to run.
        See https://docs.chalk.ai/docs/runs for more information.

        :param resolver_fqn: The fully qualified name of the resolver to trigger.
        :param context: The environment under which you'd like to query your data.

        :return: Status of the resolver run and the run ID.
        """
        ...

    def get_run_status(self, run_id: str) -> ResolverRunResponse:
        """
        Retrieves the status of a resolver run.
        See https://docs.chalk.ai/docs/runs for more information.

        :param run_id: ID of the resolver run to check.

        :return: Status of the resolver run and the run ID.
        """
        ...

    def whoami(self) -> WhoAmIResponse:
        """
        Checks the identity of your client.
        Useful as a sanity test of your configuration.

        :return: the identity of your client
        """
        ...

    def query(
        self,
        input: Mapping[Union[str, Feature, FeatureWrapper, Any], Any],
        output: Sequence[Union[str, Feature, FeatureWrapper, Any]],
        staleness: Optional[Mapping[Union[str, Feature, FeatureWrapper, Any], str]] = None,
        context: Optional[OnlineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        correlation_id: Optional[str] = None,
        query_name: Optional[str] = None,
        meta: Optional[Mapping[str, str]] = None,
    ) -> OnlineQueryResponseProtocol:
        """
        Compute features values using online resolvers.
        See https://docs.chalk.ai/docs/query-basics for more information.

        :param input: The features for which there are known values, mapped to those values.
        :param output: Outputs are the features that you'd like to compute from the inputs.
        :param staleness: Maximum staleness overrides for any output features or intermediate features.
                          See https://docs.chalk.ai/docs/query-caching for more information.
        :param context: The context object controls the environment and tags
                        under which a request should execute resolvers.
        :param preview_deployment_id: If specified, Chalk will route your request to the relevant preview
                                      deployment
        :param query_name: The name for class of query you're making, for example, "loan_application_model".
        :param correlation_id: A globally unique ID for the query, used alongside logs and
                               available in web interfaces. If None, a correlation ID will be
                               generated for you and returned on the response.
        :param meta: Arbitrary key:value pairs to associate with a query.

        :return: The outputs features and any query metadata, plus errors encountered while
        running the resolvers.
        """
        ...

    def upload_features(
        self,
        input: Mapping[Union[str, Feature, FeatureWrapper, Any], Any],
        context: Optional[OnlineQueryContext] = None,
        preview_deployment_id: Optional[str] = None,
        correlation_id: Optional[str] = None,
        query_name: Optional[str] = None,
        meta: Optional[Mapping[str, str]] = None,
    ) -> Optional[List[ChalkError]]:
        """
        Upload data to Chalk for use in offline resolvers or to prime a cache.

        :param input: The features for which there are known values, mapped to those values.
        :param context: The context object controls the environment and tags
                        under which a request should execute resolvers.
        :param preview_deployment_id: If specified, Chalk will route your request to the relevant preview
                                      deployment
        :param correlation_id: A globally unique ID for this operation, used alongside logs and
                               available in web interfaces. If None, a correlation ID will be
                               generated for you and returned on the response.
        :param query_name: Optionally associate this upload with a query name. See `.query` for more information.
        :param meta: Arbitrary key:value pairs to associate with a query.

        :return: The outputs features and any query metadata, plus errors encountered while
        running the resolvers.
        """
        ...

    def get_training_dataframe(
        self,
        input: Union[
            Mapping[Union[str, Feature], Sequence[Any]],
            pd.DataFrame,
            pl.DataFrame,
            DataFrame,
        ],
        input_times: Union[Sequence[datetime], datetime, None] = None,
        output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
        required_output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
        output_ts: bool = True,
        context: Optional[OfflineQueryContext] = None,
        dataset: Optional[str] = None,
        branch: Optional[str] = None,
        max_samples: Optional[int] = None,
        disallow_stale_results: bool = False,
    ) -> pd.DataFrame:
        """
        Compute feature values from the offline store.
        See https://docs.chalk.ai/docs/training-client for more information.

        :param input: The features for which there are known values.
                      It can be a mapping of features to a list of values for each feature,
                      or an existing dataframe.
                      Each element in the dataframe or list of values represents an observation in line
                      with the timestamp in `input_times`.
        :param input_times: A list of the times of the observations from `input`.
        :param output: The features that you'd like to sample, if they exist.
                       If an output feature was never computed for a sample (row) in the resulting DataFrame,
                       its value will be ``null``.
        :param required_output: The features that you'd like to sample and must exist in each resulting row.
                                Rows where a ``required_output`` was never stored in the offline store will be
                                skipped. This differs from specifying the feature in ``output``, where instead
                                the row would be included, but the feature value would be ``null``.
        :param output_ts: Whether to return the timestamp feature in a column named ``__chalk__.CHALK_TS`` in the
                          resulting DataFrame.
        :param context: The environment under which you'd like to query your data.
        :param dataset: A unique name that if provided will be used to generate and save a dataset
                        constructed from the list of features computed from the inputs
        :param max_samples: The maximum number of samples to include in the dataframe. If not specified, then
                            all samples will be returned.
        :param disallow_stale_results: Whether to disallow stale results when running the query. By default,
            query results might not reflect ingested data from cron runs, migrations, or online resolvers in the
            last 30 minutes.
            Setting this flag to ``True`` will ensure that the latest data is read when running the query; however,
            doing so will significantly worsen query performance.

        :return: A dataframe with columns equal to the names of the features in output,
                 and values representing the value of the observation for each input time.
                 The output maintains the ordering from `input`
        """
        ...

    def sample(
        self,
        output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
        required_output: Sequence[Union[str, Feature, FeatureWrapper, Any]] = (),
        output_id: bool = False,
        output_ts: bool = False,
        max_samples: Optional[int] = None,
        dataset: Optional[str] = None,
        branch: Optional[str] = None,
        context: Optional[OfflineQueryContext] = None,
        disallow_stale_results: bool = False,
    ) -> pd.DataFrame:
        """
        Get the most recent feature values from the offline store.
        See https://docs.chalk.ai/docs/training-client for more information.

        :param output: The features that you'd like to sample, if they exist.
                       If an output feature was never computed for a sample (row) in the resulting DataFrame,
                       its value will be ``null``.
        :param required_output: The features that you'd like to sample and must exist in each resulting row.
                                Rows where a ``required_output`` was never stored in the offline store will be
                                skipped. This differs from specifying the feature in ``output``, where instead
                                the row would be included, but the feature value would be ``null``.
        :param output_id: Whether to return the primary key feature in a column named ``__chalk__.__id__`` in the
                          resulting DataFrame.
        :param output_ts: Whether to return the timestamp feature in a column named ``__chalk__.CHALK_TS`` in the
                          resulting DataFrame.
        :param max_samples: The maximum number of rows to return.
        :param context: The environment under which you'd like to query your data.
        :param disallow_stale_results: Whether to disallow stale results when running the query. By default,
            query results might not reflect ingested data from cron runs, migrations, or online resolvers in the
            last 30 minutes.
            Setting this flag to ``True`` will ensure that the latest data is read when running the query; however,
            doing so will significantly worsen query performance.

        :return: A pandas dataframe with columns equal to the names of the features in output,
                 and values representing the value of the most recent observation.

        """
        ...



import datetime

from chalk.client import ChalkError, ErrorCode, ErrorCodeCategory, FeatureResult
from chalk.client.client_impl import OnlineQueryResponseImpl
from chalk.features import features, has_one


@features
class User:
    id: str
    email: str


@features
class PlaidAccount:
    user_id: str
    user: User = has_one(lambda: PlaidAccount.user_id == User.id)
    bank_name: str


data = [
    FeatureResult(
        field="plaid_account.user.email",
        value="faker@gmail.com",
        error=None,
        ts=datetime.datetime(2022, 12, 29, 22, 31, 10, tzinfo=datetime.timezone.utc),
    ),
    FeatureResult(
        field="plaid_account.user.id",
        value="IuiTTRfsldkj0878650X6Kn9mjDE73",
        error=None,
        ts=datetime.datetime(2022, 12, 29, 22, 31, 10, tzinfo=datetime.timezone.utc),
    ),
    FeatureResult(
        field="plaid_account.bank_name",
        value="Wells Fargo",
        error=None,
        ts=datetime.datetime(2022, 12, 29, 22, 31, 10, tzinfo=datetime.timezone.utc),
    ),
]

data_response = OnlineQueryResponseImpl(data=data, errors=[], warnings=[])
error_response = OnlineQueryResponseImpl(
    data=[],
    warnings=[],
    errors=[
        ChalkError(
            code=ErrorCode.INVALID_QUERY,
            category=ErrorCodeCategory.REQUEST,
            message='Could not execute query. Missing required inputs: "plaid_account.access_token"',
            exception=None,
            feature=None,
            resolver=None,
        )
    ],
)

repr_result = """                      field  ...                        ts
0  plaid_account.user.email  ... 2022-12-29 22:31:10+00:00
1     plaid_account.user.id  ... 2022-12-29 22:31:10+00:00
2   plaid_account.bank_name  ... 2022-12-29 22:31:10+00:00

[3 rows x 4 columns]"""

html_result = """<div><stylescoped>.dataframetbodytrth:only-of-type{vertical-align:middle;}.dataframetbodytrth{vertical-align:top;}.dataframetheadth{text-align:right;}</style><tableborder="1"class="dataframe"><thead><trstyle="text-align:right;"><th></th><th>field</th><th>value</th><th>error</th><th>ts</th></tr></thead><tbody><tr><th>0</th><td>plaid_account.user.email</td><td>faker@gmail.com</td><td>None</td><td>2022-12-2922:31:10+00:00</td></tr><tr><th>1</th><td>plaid_account.user.id</td><td>IuiTTRfsldkj0878650X6Kn9mjDE73</td><td>None</td><td>2022-12-2922:31:10+00:00</td></tr><tr><th>2</th><td>plaid_account.bank_name</td><td>WellsFargo</td><td>None</td><td>2022-12-2922:31:10+00:00</td></tr></tbody></table></div>"""

error_repr = """                      code                   category  ... feature resolver
0  ErrorCode.INVALID_QUERY  ErrorCodeCategory.REQUEST  ...    None     None

[1 rows x 6 columns]"""

error_html = '<div><stylescoped>.dataframetbodytrth:only-of-type{vertical-align:middle;}.dataframetbodytrth{vertical-align:top;}.dataframetheadth{text-align:right;}</style><tableborder="1"class="dataframe"><thead><trstyle="text-align:right;"><th></th><th>code</th><th>category</th><th>message</th><th>exception</th><th>feature</th><th>resolver</th></tr></thead><tbody><tr><th>0</th><td>ErrorCode.INVALID_QUERY</td><td>ErrorCodeCategory.REQUEST</td><td>Couldnotexecutequery.Missingrequiredinpu...</td><td>None</td><td>None</td><td>None</td></tr></tbody></table></div>'


class TestClientStringRepresentation:
    def test_OQRW_data_repr(self):
        assert "".join(repr_result.split()) == "".join(repr(data_response).split())

    def test_OQRW_data_repr_html(self):
        assert "".join(html_result.split()) == "".join(data_response._repr_html_().split())

    def test_OQRW_error_repr(self):
        assert "".join(error_repr.split()) == "".join(repr(error_response).split())

    def test_OQRW_error_repr_html(self):
        assert "".join(error_html.split()) == "".join(error_response._repr_html_().split())



from sqlalchemy import text

from chalk.features import features, unwrap_feature
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.sql.finalized_query import FinalizedChalkQuery, Finalizer


@features
class ColNameMappingFeatures:
    id: str
    foo: str
    bar: str


def test_get_col_name_mapping():
    finalized_query = FinalizedChalkQuery(
        query=text("select 1"),
        params={},
        finalizer=Finalizer.ALL,
        incremental_settings=None,
        source=BaseSQLSource(name=""),
        fields={
            "id": unwrap_feature(ColNameMappingFeatures.id),
            "foo": unwrap_feature(ColNameMappingFeatures.foo),
            "bar": unwrap_feature(ColNameMappingFeatures.bar),
        },
    )
    col_name_mapping = finalized_query.get_col_name_mapping(result_columns=["ID", "foo", "bAr"], expected_features=None)
    assert col_name_mapping == {
        "ID": "col_name_mapping_features.id",
        "foo": "col_name_mapping_features.foo",
        "bAr": "col_name_mapping_features.bar",
    }


def test_get_col_name_mapping_with_expected():
    finalized_query = FinalizedChalkQuery(
        query=text("select 1"),
        params={},
        finalizer=Finalizer.ALL,
        incremental_settings=None,
        source=BaseSQLSource(name=""),
        fields={},
    )
    col_name_mapping = finalized_query.get_col_name_mapping(
        result_columns=["ID", "foo", "bAr"],
        expected_features=[
            unwrap_feature(ColNameMappingFeatures.id),
            unwrap_feature(ColNameMappingFeatures.foo),
            unwrap_feature(ColNameMappingFeatures.bar),
        ],
    )
    assert col_name_mapping == {
        "ID": "col_name_mapping_features.id",
        "foo": "col_name_mapping_features.foo",
        "bAr": "col_name_mapping_features.bar",
    }



import inspect
from typing import Any, Callable, List, Optional, Type, TypeVar, Union

from pydantic import BaseModel
from typing_extensions import ParamSpec

from chalk.state import KeyedState
from chalk.streams._file_source import FileSource
from chalk.streams._kafka_source import KafkaSource
from chalk.streams._windows import Windowed, get_duration_secs, get_name_with_duration, windowed
from chalk.streams.base import StreamSource
from chalk.utils import MachineType

__all__ = [
    "FileSource",
    "KafkaSource",
    "KeyedState",
    "StreamSource",
    "Windowed",
    "stream",
    "windowed",
    "get_name_with_duration",
    "get_duration_secs",
]

P = ParamSpec("P")
T = TypeVar("T")

MessageType = TypeVar("MessageType", bound=BaseModel)


def stream(
    *,
    source: StreamSource,
    mode: Optional[str] = None,
    environment: Optional[Union[List[str], str]] = None,
    machine_type: Optional[MachineType] = None,
    message: Optional[Type[Any]] = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    caller_frame = inspect.stack()[1]
    caller_filename = caller_frame.filename
    caller_globals = caller_frame.frame.f_globals
    caller_locals = caller_frame.frame.f_locals
    from chalk.features.resolver import parse_and_register_stream_resolver

    def decorator(fn: Callable[P, T]) -> Callable[P, T]:
        return parse_and_register_stream_resolver(
            caller_globals=caller_globals,
            caller_locals=caller_locals,
            fn=fn,
            source=source,
            mode=mode,
            caller_filename=caller_filename,
            environment=environment,
            machine_type=machine_type,
            message=message,
        )

    return decorator



import importlib
import importlib.util
import os
import sys
import traceback
from pathlib import Path
from typing import Callable, List, Optional

from chalk.features.resolver import Resolver
from chalk.gitignore.gitignore_parser import parse_gitignore
from chalk.parsed.duplicate_input_gql import FailedImport
from chalk.sql._internal.sql_file_resolver import ResolverResult, get_sql_file_resolvers
from chalk.sql._internal.sql_source import BaseSQLSource
from chalk.utils.log_with_context import get_logger
from chalk.utils.paths import get_directory_root

_logger = get_logger(__name__)


def py_path_to_module(path: Path, repo_root: Path) -> str:
    try:
        p = path.relative_to(repo_root)
    except ValueError:
        p = path
    return str(p)[: -len(".py")].replace("./", "").replace("/", ".")


def _import_all_files() -> List[FailedImport]:
    project_root: Optional[Path] = get_directory_root()
    if project_root is None:
        return [
            FailedImport(
                filename="",
                module="",
                traceback="Could not find chalk.yaml in this directory or any parent directory",
            )
        ]
    sql_resolver_results: List[ResolverResult] = list(get_sql_file_resolvers(project_root, BaseSQLSource.registry))
    failed_imports: List[FailedImport] = import_sql_file_resolvers(project_root)
    failed_imports.extend(import_all_python_files_from_dir(project_root=project_root))
    return failed_imports


def import_sql_file_resolvers(path: Path):
    sql_resolver_results: List[ResolverResult] = list(get_sql_file_resolvers(path, BaseSQLSource.registry))
    failed_imports: List[FailedImport] = []
    for result in sql_resolver_results:
        if result.resolver:
            Resolver.registry.append(result.resolver)
            if Resolver.hook:
                Resolver.hook(result.resolver)
        if result.errors:
            for error in result.errors:
                failed_imports.append(
                    FailedImport(
                        traceback=error.display,
                        filename=error.path,
                        module=error.path,
                    )
                )
    return failed_imports


def _search_recursively_for_file(base: Path, filename: str) -> List[Path]:
    ans = []
    assert base.is_dir()
    while True:
        filepath = base / filename
        if filepath.exists():
            ans.append(filepath)
        parent = base.parent
        if parent == base:
            return ans
        base = parent


def _is_relevant(x: Path, matching_functions: List[Callable[[Path], bool]]):
    return all((not match(x) for match in matching_functions))


def import_all_python_files_from_dir(project_root: Path) -> List[FailedImport]:
    project_root = project_root.absolute()
    matching_functions: List[Callable[[Path], bool]] = []
    matching_functions.extend(parse_gitignore(str(x)) for x in _search_recursively_for_file(project_root, ".gitignore"))
    matching_functions.extend(
        parse_gitignore(str(x)) for x in _search_recursively_for_file(project_root, ".chalkignore")
    )
    matching_functions.append(lambda x: x.name == "setup.py")

    cwd = os.getcwd()
    os.chdir(project_root)
    repo_root = Path(project_root)

    # If we don't import both of these, we get in trouble.
    sys.path.append(str(repo_root.resolve()))
    sys.path.append(str(repo_root.parent.resolve()))

    repo_files = sorted({p.resolve() for p in repo_root.glob("**/*.py") if p.is_file()})

    _logger.debug(f"REPO_ROOT: {repo_root.resolve()}")
    # _logger.debug(f"REPO_FILES: {repo_files}")

    venv = os.environ.get("VIRTUAL_ENV")
    module_paths = [
        (py_path_to_module(repo_file, repo_root), repo_file)
        for repo_file in repo_files
        if venv is None or Path(venv) not in repo_file.parents
    ]

    errors: List[FailedImport] = []
    for (module_path, filename) in module_paths:
        if (
            module_path.startswith(".eggs")
            or module_path.startswith("venv")
            or not _is_relevant(filename, matching_functions)
        ):
            continue

        try:
            importlib.import_module(module_path)
        except Exception:
            filename = filename.resolve()
            # relevant_file = any("from chalk." or "import chalk." in c for c in f)
            # if relevant_file:
            ex_type, ex_value, ex_traceback = sys.exc_info()
            tb = traceback.extract_tb(ex_traceback)
            line = 0
            for i, l in enumerate(tb):
                if filename == Path(l.filename).resolve():
                    line = i
                    break

            relevant_traceback = f"""Exception in module {module_path}:
{os.linesep.join(traceback.format_tb(ex_traceback)[line:])}
\n{ex_type and ex_type.__name__}: {str(ex_value)}
"""
            errors.append(
                FailedImport(
                    traceback=relevant_traceback,
                    filename=str(filename),
                    module=module_path,
                )
            )

            _logger.debug(f"Failed while importing {module_path}", exc_info=True)
            continue

    os.chdir(cwd)
    return errors



import dataclasses
from datetime import timedelta
from typing import Optional


@dataclasses.dataclass
class IncrementalSettings:
    mode: str = "row"  # "row" or "group" or "parameter"
    lookback_period: Optional[timedelta] = None
    incremental_column: Optional[str] = None



from __future__ import annotations

import uuid
from datetime import datetime
from enum import Enum
from typing import Any, Dict, List, Mapping, Optional, Union

import numpy as np
from pydantic import BaseModel, Field, root_validator


class OnlineQueryContext(BaseModel):
    # The environment under which to run the resolvers.
    # API tokens can be scoped to an # environment.
    # If no environment is specified in the query,
    # but the token supports only a single environment,
    # then that environment will be taken as the scope
    # for executing the request.
    environment: Optional[str] = None

    # The tags used to scope the resolvers.
    # More information at https://docs.chalk.ai/docs/resolver-tags
    tags: Optional[List[str]] = None


class OfflineQueryContext(BaseModel):
    # The environment under which to run the resolvers.
    # API tokens can be scoped to an # environment.
    # If no environment is specified in the query,
    # but the token supports only a single environment,
    # then that environment will be taken as the scope
    # for executing the request.
    environment: Optional[str] = None


class ErrorCode(str, Enum):
    # The query contained features that do not exist.
    PARSE_FAILED = "PARSE_FAILED"

    # A resolver was required as part of running the dependency
    # graph that could not be found.
    RESOLVER_NOT_FOUND = "RESOLVER_NOT_FOUND"

    # The query is invalid. All supplied features need to be
    # rooted in the same top-level entity.
    INVALID_QUERY = "INVALID_QUERY"

    # A feature value did not match the expected schema
    # (eg. `incompatible type "int"; expected "str"`)
    VALIDATION_FAILED = "VALIDATION_FAILED"

    # The resolver for a feature errored.
    RESOLVER_FAILED = "RESOLVER_FAILED"

    # A crash in a resolver that was to produce an input for
    # the resolver crashed, and so the resolver could not run
    # crashed, and so the resolver could not run.
    UPSTREAM_FAILED = "UPSTREAM_FAILED"

    # The request was submitted with an invalid authentication header.
    UNAUTHENTICATED = "UNAUTHENTICATED"

    # An unspecified error occurred.
    INTERNAL_SERVER_ERROR = "INTERNAL_SERVER_ERROR"

    @classmethod
    def category(cls, c: Union[ErrorCode, str]) -> ErrorCodeCategory:
        c = ErrorCode[c]
        return {
            cls.PARSE_FAILED: ErrorCodeCategory.REQUEST,
            cls.RESOLVER_NOT_FOUND: ErrorCodeCategory.REQUEST,
            cls.INVALID_QUERY: ErrorCodeCategory.REQUEST,
            cls.VALIDATION_FAILED: ErrorCodeCategory.FIELD,
            cls.RESOLVER_FAILED: ErrorCodeCategory.FIELD,
            cls.UPSTREAM_FAILED: ErrorCodeCategory.FIELD,
            cls.UNAUTHENTICATED: ErrorCodeCategory.NETWORK,
            cls.INTERNAL_SERVER_ERROR: ErrorCodeCategory.NETWORK,
        }[c]


class ErrorCodeCategory(str, Enum):

    # Request errors are raised before execution of your
    # resolver code. They may occur due to invalid feature
    # names in the input or a request that cannot be satisfied
    # by the resolvers you have defined.
    REQUEST = "REQUEST"

    # Field errors are raised while running a feature resolver
    # for a particular field. For this type of error, you’ll
    # find a feature and resolver attribute in the error type.
    # When a feature resolver crashes, you will receive null
    # value in the response. To differentiate from a resolver
    # returning a null value and a failure in the resolver,
    # you need to check the error schema.
    FIELD = "FIELD"

    # Network errors are thrown outside your resolvers.
    # For example, your request was unauthenticated,
    # connection failed, or an error occurred within Chalk.
    NETWORK = "NETWORK"


class ChalkException(BaseModel):
    # The name of the class of the exception.
    kind: str

    # The message taken from the exception.
    message: str

    # The stacktrace produced by the code.
    stacktrace: str

    class Config:
        frozen = True


class ChalkError(BaseModel):
    # The type of the error
    code: ErrorCode

    # The category of the error, given in the type field for the error codes.
    # This will be one of "REQUEST", "NETWORK", and "FIELD".
    category: ErrorCodeCategory = ErrorCodeCategory.NETWORK

    # A readable description of the error message.
    message: str

    # The exception that caused the failure, if applicable.
    exception: Optional[ChalkException] = None

    # The fully qualified name of the failing feature, eg.user.identity.has_voip_phone.
    feature: Optional[str] = None

    # The fully qualified name of the failing resolver, eg.my.project.get_fraud_score.
    resolver: Optional[str] = None

    class Config:
        frozen = True

    def copy_for_feature(self, feature: str) -> "ChalkError":
        return self.copy(update={"feature": feature})

    @root_validator
    def validate_category(cls, values: Dict[str, Any]):
        values["category"] = ErrorCode.category(values["code"])
        return values


class ResolverRunStatus(str, Enum):
    RECEIVED = "received"
    SUCCEEDED = "succeeded"
    FAILED = "failed"


class ResolverRunResponse(BaseModel):
    id: str
    status: ResolverRunStatus


class WhoAmIResponse(BaseModel):
    user: str


class FeatureResolutionMeta(BaseModel):
    chosen_resolver_fqn: str
    cache_hit: bool

    class Config:
        frozen = True


class FeatureResult(BaseModel):
    # The name of the feature requested, eg.user.identity.has_voip_phone.
    field: str

    # The value of the requested feature.
    # If an error was encountered in resolving this feature,
    # this field will be empty.
    value: Any

    pkey: Any

    # The error code encountered in resolving this feature.
    # If no error occurred, this field is empty.
    error: Optional[ChalkError] = None

    # The time at which this feature was computed.
    # This value could be significantly in the past if you're using caching.
    ts: Optional[datetime] = None

    meta: Optional[FeatureResolutionMeta] = None


class ExchangeCredentialsRequest(BaseModel):
    client_id: str
    client_secret: str
    grant_type: str


class ExchangeCredentialsResponse(BaseModel):
    access_token: str
    token_type: str
    expires_in: int
    # expires_at: datetime
    api_server: str
    primary_environment: Optional[str] = None
    engines: Optional[Mapping[str, str]] = None


class OfflineQueryInput(BaseModel):
    columns: List[str]
    values: List[List[Any]]


class OnlineQueryRequest(BaseModel):
    inputs: Mapping[str, Any]
    outputs: List[str]
    staleness: Optional[Mapping[str, str]] = None
    context: Optional[OnlineQueryContext] = None
    include_meta: bool = False
    skip_online_storage: Optional[bool] = False
    skip_offline_storage: Optional[bool] = False
    skip_metrics_storage: Optional[bool] = False
    skip_cache_lookups: Optional[bool] = False
    correlation_id: Optional[str] = None
    query_name: Optional[str] = None
    deployment_id: Optional[str] = None
    meta: Optional[Mapping[str, str]] = None


class TriggerResolverRunRequest(BaseModel):
    resolver_fqn: str


class QueryMeta(BaseModel):
    execution_duration_s: float
    deployment_id: Optional[str] = None
    query_id: Optional[str] = None


class OnlineQueryResponse(BaseModel):
    data: List[FeatureResult]
    errors: Optional[List[ChalkError]] = None
    meta: Optional[QueryMeta] = None

    def for_fqn(self, fqn: str):
        return next((x for x in self.data if x.field == fqn), None)

    class Config:
        json_encoders = {
            np.integer: int,
            np.floating: float,
        }


class CreateOfflineQueryJobRequest(BaseModel):
    output: List[str] = Field(description="A list of output feature root fqns to query")
    required_output: List[str] = Field(default_factory=list, description="A list of required output feature root fqns")
    destination_format: str = Field(description="The desired output format. Should be 'CSV' or 'PARQUET'")
    job_id: Optional[uuid.UUID] = Field(
        default=None,
        description=(
            "A unique job id. If not specified, one will be auto generated by the server. If specified by the client, "
            "then jobs with the same ID will be rejected."
        ),
    )
    input: Optional[OfflineQueryInput] = Field(default=None, description="Any givens")
    max_samples: Optional[int] = Field(
        default=None,
        description="The maximum number of samples. If None, no limit",
    )
    max_cache_age_secs: Optional[int] = Field(
        default=None,  # Defaults to ``OFFLINE_QUERY_MAX_CACHE_AGE_SECS`` in the chalkengine config
        description=(
            "The maximum staleness, in seconds, for how old the view on the offline store can be. That is, "
            "data ingested within this interval will not be reflected in this offline query. "
            "Set to ``0`` to ignore the cache. If not specified, it defaults to 30 minutes."
        ),
    )
    observed_at_lower_bound: Optional[datetime] = Field(
        default=None,
        description="The lower bound for the observed at timestamp (inclusive). If not specified, defaults to the beginning of time",
    )
    observed_at_upper_bound: Optional[datetime] = Field(
        default=None,
        description="The upper bound for the observed at timestamp (inclusive). If not specified, defaults to the end of time.",
    )
    dataset_name: Optional[str] = None
    branch: Optional[str] = None


# FIXME document
class ComputeResolverOutputRequest(BaseModel):
    input: OfflineQueryInput
    resolver_fqn: str
    branch: Optional[str] = None
    environment: Optional[str] = None


class RecomputeResolverOutputRequest(BaseModel):
    persistent_id: str
    resolver_fqn: str
    branch: Optional[str] = None
    environment: Optional[str] = None


class ComputeResolverOutputResponse(BaseModel):
    job_id: str
    persistent_id: str
    errors: Optional[List[ChalkError]] = None


class OfflineQueryRequest(BaseModel):
    """V1 OfflineQueryRequest. Not used by the current Chalk Client."""

    output: List[str]  # output features which can be null
    input: Optional[OfflineQueryInput] = None
    dataset: Optional[str] = None
    max_samples: Optional[int] = None
    max_cache_age_secs: Optional[int] = None
    required_outputs: List[str] = Field(default_factory=list)  # output features which cannot be null


class OfflineQueryResponse(BaseModel):
    """V1 OfflineQueryResponse. Not used by the current Chalk Client."""

    columns: List[str]
    output: List[List[Any]]
    errors: Optional[List[ChalkError]] = None


class CreateOfflineQueryJobResponse(BaseModel):
    """
    Attributes:
        job_id: A job ID, which can be used to retrieve the results.
    """

    job_id: uuid.UUID
    version: int = 1  # Field is deprecated
    errors: Optional[List[ChalkError]] = None


class GetOfflineQueryJobResponse(BaseModel):
    is_finished: bool = Field(description="Whether the export job is finished (it runs asynchronously)")
    version: int = Field(
        default=1,  # Backwards compatibility
        description=(
            "Version number representing the format of the data. The client uses this version number "
            "to properly decode and load the query results into DataFrames."
        ),
    )
    urls: List[str] = Field(
        description="A list of short-lived, authenticated URLs that the client can download to retrieve the exported data."
    )
    errors: Optional[List[ChalkError]] = None



from chalk.features import Features, features


@features
class UserProfile:
    user_id: str
    profile_id: str
    address: int


@features
class Account:
    account_id: str
    balance: int


@features
class Transaction:
    user_id: str
    amount: int
    source_account_id: int
    dest_account_id: int


@features
class MypyUserFeatures:
    uid: str
    name: str
    bday: str
    age: int


def name_age_good(
    age: MypyUserFeatures.age,
) -> Features[MypyUserFeatures.name, MypyUserFeatures.age]:
    # Correctly knows that age is an in
    age_int_cast: int = age
    print(age_int_cast)
    # Correctly casts UserFeatures.__init__ to Features[name, age]
    return MypyUserFeatures(name="4", age=4)


def name_age_good_flipped(
    age: MypyUserFeatures.age,
) -> Features[MypyUserFeatures.age, MypyUserFeatures.name]:
    # Flips name and age to match
    return MypyUserFeatures(name="4", age=4)


def name_age_bad() -> Features[MypyUserFeatures.name, MypyUserFeatures.age]:
    # We're getting Features[name, bday] and expected Features[name, age]
    return MypyUserFeatures(name="4", bday="")


def name_bad_type() -> Features[MypyUserFeatures.name]:
    # Knows that name should be a string, not an int
    return MypyUserFeatures(name=4)  # pyright: ignore



from datetime import datetime
from typing import Optional

import pytest

from chalk.features import feature, feature_time, features, has_one
from chalk.importer import import_sql_file_resolvers

fraud_mysql = None
risk_sf = None
marketing_sf = None

try:
    import pymysql
    import snowflake

    from chalk.sql import MySQLSource, SnowflakeSource
    from chalk.sql._internal.sql_file_resolver import *

    fraud_mysql = MySQLSource(name="FRAUD_MYSQL")
    risk_sf = SnowflakeSource(name="RISK_SF")
    marketing_sf = SnowflakeSource(name="MARKETING_SF")
except ImportError:
    snowflake = None
    pymysql = None


@features
class Burrito:
    id: str
    name: str
    size: str


@features
class Meat:
    id: int
    name: str
    calories: float

    burrito_id: str
    burrito: Burrito = has_one(lambda: Burrito.id == Meat.burrito_id)


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_simple_sql_file_resolver():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.function_definition == "select name, size from burritos_table where id = ${burrito.id}"
    assert res.inputs == [Burrito.id]
    assert res.output.__columns__ == (Burrito.name, Burrito.size)
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id"}
    assert (
        str(res.fn()._query)
        == """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select name, size from burritos_table where id = :__chalk_burrito_id__"""
    )


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_multiple_input_sql_file_resolver():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select name, size from burritos_table where id = ${burrito.id} and name=${burrito.name}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert Burrito.id._chalk_feature in res.inputs and Burrito.name._chalk_feature in res.inputs
    assert res.output.__columns__ == (Burrito.name, Burrito.size)
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id", "__chalk_burrito_name__": "burrito.name"}
    assert res.default_args == [ResolverArgErrorHandler(None)] * 2


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_incremental_sql_file_resolver():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        -- type: offline
        -- incremental:
        --    mode: parameter
        --    lookback_period: 60m
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Burrito.id]
    assert res.output.__columns__ == (Burrito.name, Burrito.size)
    assert isinstance(res, OfflineResolver)
    assert res.fn().incremental_settings.mode == "parameter"
    assert str(res.fn().incremental_settings.lookback_period) == "1:00:00"
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id"}


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_no_sql_source():
    sources = [marketing_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert resolver_result.errors[0].display == "Source RISK_SF not found"


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_two_sql_sources():
    sources = [marketing_sf, risk_sf]
    sql_string = """-- source: snowflake
        -- cron: 1h
        -- resolves: burrito
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert resolver_result.errors[0].display.startswith("More than one snowflake source exists")


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_comments_parsed():
    sources = [marketing_sf]
    sql_string = """-- Here is a normal comment
        -- source: snowflake
        -- cron: 1h
        -- resolves: burrito
        -- incremental:
        --    mode: parameter
        --    lookback_period: 60m
        -- tags:
        --    - one:1
        --    - two:2
        -- Here is another comment with a colon : in : it
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert len(resolver_result.errors) == 0
    assert resolver_result.resolver.doc == "Here is a normal comment\nHere is another comment with a colon : in : it"
    assert resolver_result.resolver.tags == ["one:1", "two:2"]


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_single_tag_parse():
    sources = [marketing_sf]
    sql_string = """-- Here is a normal comment
        -- source: snowflake
        -- cron: 1h
        -- resolves: burrito
        -- incremental:
        --    mode: parameter
        --    lookback_period: 60m
        -- tags: singletag
        -- Here is another comment with a colon : in : it
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert len(resolver_result.errors) == 0
    assert resolver_result.resolver.tags == ["singletag"]


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_single_tag_malformed_parse():
    sources = [marketing_sf]
    sql_string = """-- Here is a normal comment
        -- source: snowflake
        -- cron: 1h
        -- resolves: burrito
        -- incremental:
        --    mode: parameter
        --    lookback_period: 60m
        -- tags:
        --    one: 1
        --    two: 2
        -- Here is another comment with a colon : in : it
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert len(resolver_result.errors) == 1
    assert resolver_result.errors[0].display == "Parameter 'tags' value {'one': 1, 'two': 2} should be a list"


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_malformed_sql_query():
    sources = [marketing_sf, risk_sf]
    sql_string = """-- source: bigquery
        -- cron: 1h
        -- resolves: burrito
        -- bad comment
        -- another bad comment
        select name, size from burritos_table whereASDFW id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert len(resolver_result.errors) == 1
    assert resolver_result.errors[0].display.startswith("Cannot SQL parse")


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_missing_required_source():
    sources = [marketing_sf, risk_sf]
    sql_string = """-- missing: required_comment
        -- cron: 1h
        -- resolves: burrito
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert (
        resolver_result.errors[0].display == "Missing value for 'source'. Please specify via "
        "source type in ['snowflake', 'postgres', 'postgresql', 'mysql', "
        "'bigquery', 'cloudsql', 'redshift', 'sqlite'], or via source name in "
        "['MARKETING_SF', 'RISK_SF']"
    )


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_missing_required_resolves():
    sources = [marketing_sf]
    sql_string = """-- missing: required_comment
        -- source: MARKETING_SF
        -- cron: 1h
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert resolver_result.errors[0].display.startswith("Key 'resolves' must be described in comments.")


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_default_variable_sql_resolver():
    sources = [risk_sf]
    sql_string1 = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select b_name as name, b_size as size
        from burritos_table where id = ${burrito.id | "b_1"}"""
    sql_string2 = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select b_name as name, b_size as size
        from burritos_table where id = ${burrito.id or "b_1"}"""
    sql_string_result1 = SQLStringResult(path="sample_path", sql_string=sql_string1, error=None)
    sql_string_result2 = SQLStringResult(path="sample_path", sql_string=sql_string2, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result1)
    resolver_result2 = get_sql_file_resolver(sources, sql_string_result2)
    assert resolver_result == resolver_result2
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Burrito.id]
    assert res.output.__columns__ == (Burrito.name, Burrito.size)
    assert res.default_args == [ResolverArgErrorHandler("b_1")]
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id"}
    assert (
        str(res.fn()._query)
        == """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select b_name as name, b_size as size
        from burritos_table where id = :__chalk_burrito_id__"""
    )


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_default_variable_sql_resolver_int():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: meat
        select id, name
        from meat_table where id = ${meat.id | 2}"""

    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Meat.id]
    assert res.output.__columns__ == (Meat.id, Meat.name)
    assert res.default_args == [ResolverArgErrorHandler(2)]
    assert (
        str(res.fn()._query)
        == """-- source: RISK_SF
        -- cron: 1h
        -- resolves: meat
        select id, name
        from meat_table where id = :__chalk_meat_id__"""
    )


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_default_variable_sql_resolver_float():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: meat
        select id, name
        from meat_table where calories = ${meat.calories | 2.2}"""

    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Meat.calories]
    assert res.output.__columns__ == (Meat.id, Meat.name)
    assert res.default_args == [ResolverArgErrorHandler(2.2)]


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_malformed_default_variable_sql_resolver():
    sources = [risk_sf]
    sql_string1 = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select b_name as name, b_size as size
        from burritos_table where id = ${burrito.id | burrito.size | burrito.value}"""
    sql_string_result1 = SQLStringResult(path="sample_path", sql_string=sql_string1, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result1)
    assert resolver_result.errors[0].display.startswith("If character '|' is used, both variable name")


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_as_sql_resolver():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select b_name as name, b_size as size
        from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Burrito.id]
    assert res.output.__columns__ == (Burrito.name, Burrito.size)
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id"}


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_sql_resolver_with_table_names():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select burritos_table.name, burritos_table.size
        from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Burrito.id]
    assert res.output.__columns__ == (Burrito.name, Burrito.size)
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id"}


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_as_sql_resolver_with_table_names():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        select burritos_table.name as name, burritos_table.size as size
        from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Burrito.id]
    assert res.output.__columns__ == (Burrito.name, Burrito.size)
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id"}


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_one_sql_file_resolver():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- count: one
        -- resolves: burrito
        select name, size from burritos_table where id = ${burrito.id}"""
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.function_definition == "select name, size from burritos_table where id = ${burrito.id}"
    assert res.inputs == [Burrito.id]
    assert res.output.features == (Burrito.name, Burrito.size)
    assert resolver_result.fields == {"name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_burrito_id__": "burrito.id"}
    assert res.fn().finalizer == "OneOrNone"


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_sql_join_hop():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: burrito
        SELECT burritos_table.id AS id,
        burritos_table.name AS name,
        burritos_table.size AS size,
        FROM burritos_table JOIN meat_table ON burritos_table.id = meat_table.burrito_id
        WHERE meat_table.id = ${meat.id}"""
    # in this case, 'id = ${}' wouldn't work! Be careful!
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Meat.id]
    assert res.output.__columns__ == (Burrito.id, Burrito.name, Burrito.size)
    assert resolver_result.fields == {"id": "burrito.id", "name": "burrito.name", "size": "burrito.size"}
    assert resolver_result.args == {"__chalk_meat_id__": "meat.id"}


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_sql_join_stay():
    sources = [risk_sf]
    sql_string = """-- source: RISK_SF
        -- cron: 1h
        -- resolves: meat
        SELECT meat_table.name AS name, meat_table.id as id
        FROM burritos_table JOIN meat_table ON burritos_table.id = meat_table.burrito_id
        WHERE meat_table.id = ${meat.id}"""
    # in this case, 'id = ${}' wouldn't work! Be careful!
    sql_string_result = SQLStringResult(path="sample_path", sql_string=sql_string, error=None)
    resolver_result = get_sql_file_resolver(sources, sql_string_result)
    assert not resolver_result.errors
    res = resolver_result.resolver
    assert res.inputs == [Meat.id]
    assert res.output.__columns__ == (Meat.name, Meat.id)
    assert resolver_result.fields == {"name": "meat.name", "id": "meat.id"}
    assert resolver_result.args == {"__chalk_meat_id__": "meat.id"}


@pytest.mark.skipif(snowflake is None, reason="Snowflake is not installed")
def test_find_sql_resolver_files():
    project_root = Path(".")
    failed_imports = import_sql_file_resolvers(project_root)
    assert not failed_imports



import pytest

from chalk.client.dataset import DatasetVersion, load_dataset
from chalk.features import features


@pytest.mark.skip("not implemented fully")
def test_load_dataset():
    @features
    class CRUser:
        id: int
        favorite_color: str
        favorite_number: int

    data = load_dataset(
        uris=["gs://compute-resolver-results/1673236985507958784_b8e031b8-d7a1-4dc9-beef-b5837889a2d9.parquet"],
        version=DatasetVersion.COMPUTE_RESOLVER_OUTPUT_V1,
    )

    print(data)
    assert data == 1



from typing import cast

import pytest
from sqlalchemy import text

from chalk import realtime
from chalk.features import DataFrame, features
from chalk.sql import BaseSQLSourceProtocol
from chalk.testing import assert_frame_equal_ignore_order

try:
    import polars as pl
except ImportError:
    pl = None

seed_db = """
DROP TABLE IF EXISTS mappings_table;

CREATE TABLE mappings_table (
    id INT PRIMARY KEY,
    ufrom TEXT,   -- maps to `u_from` feature, implicitly
    uto TEXT,     -- maps to `u_to` feature, implicitly
    col_a TEXT,   -- maps to `column_a` feature, explicitly
    swap_b TEXT,  -- maps to `swap_c` feature, explicitly
    swap_c TEXT,  -- maps to `swap_b` feature, explicitly
    unused TEXT   -- unused
);

INSERT INTO mappings_table(id, ufrom, uto, col_a, swap_b, swap_c, unused) VALUES
  (0, 'from', 'to', 'value_a', 'value_b', 'value_c', 'unused');
"""


@pytest.fixture
def init_db(temp_sqlite_source: BaseSQLSourceProtocol):
    eng = temp_sqlite_source.get_engine()
    with eng.connect() as cnx, cnx.begin():
        for statement in seed_db.split(";"):
            if statement != "":
                cnx.execute(text(statement))


@features
class MappingFeatures:
    id: int
    u_from: str
    u_to: str
    column_a: str
    swap_b: str
    swap_c: str


@pytest.mark.skipif(pl is None, reason="Polars is not installed")
def test_sql_resolver_implicit_mapping_rules(init_db: None, temp_sqlite_source: BaseSQLSourceProtocol):
    @realtime
    def resolve_mapping_features() -> DataFrame[MappingFeatures]:
        return cast(
            DataFrame,
            temp_sqlite_source.query_string(
                """
            SELECT * FROM mappings_table;
            """,
                fields={
                    "col_a": MappingFeatures.column_a,
                    "swap_b": MappingFeatures.swap_c,
                    "swap_c": MappingFeatures.swap_b,
                },
            ).all(),
        )

    results = resolve_mapping_features()
    assert_frame_equal_ignore_order(
        results,
        DataFrame(
            [MappingFeatures(id=0, u_from="from", u_to="to", column_a="value_a", swap_b="value_c", swap_c="value_b")]
        ),
    )



from os import PathLike
from typing import Any, Dict, Optional, TypeVar, Union

from sqlalchemy.engine.url import URL

from chalk.sql._internal.sql_source import BaseSQLSource, TableIngestMixIn
from chalk.sql.protocols import SQLSourceWithTableIngestProtocol

T = TypeVar("T")


class SQLiteInMemorySourceImpl(TableIngestMixIn, BaseSQLSource, SQLSourceWithTableIngestProtocol):
    def __init__(self, name: Optional[str] = None):
        self.ingested_tables: Dict[str, Any] = {}
        BaseSQLSource.__init__(self, name=name)

    def local_engine_url(self) -> URL:
        return URL.create(drivername="sqlite", database=":memory:", query={"check_same_thread": "true"})


class SQLiteFileSourceImpl(TableIngestMixIn, BaseSQLSource, SQLSourceWithTableIngestProtocol):
    def __init__(self, filename: Union[PathLike, str], name: Optional[str] = None):
        self.filename = str(filename)
        self.ingested_tables: Dict[str, Any] = {}
        BaseSQLSource.__init__(self, name=name)

    def local_engine_url(self) -> URL:
        return URL.create(drivername="sqlite", database=self.filename, query={"check_same_thread": "true"})



import json

import pytest

from chalk.client.models import ChalkError, ErrorCode, ErrorCodeCategory, OnlineQueryResponse


@pytest.mark.parametrize("code", list(ErrorCode))
def test_create_chalk_error(code: ErrorCode):
    assert isinstance(ErrorCode.category(code), ErrorCodeCategory)
    error = ChalkError(code=code, message="")
    assert error.category == ErrorCode.category(error.code)
    assert error.dict()["category"] == ErrorCode.category(error.code)


def test_chalk_error_serialization():
    x = ChalkError(code=ErrorCode.RESOLVER_FAILED, message="Test error")
    x_jsoned = x.json()
    x_dict = json.loads(x_jsoned)
    del x_dict["category"]
    x_parsed = ChalkError(**x_dict)
    assert x == x_parsed



import enum
import logging
import os
import pathlib
import uuid
from datetime import date, datetime, timezone
from typing import Any, List, Tuple, Type, cast

import pytest
from sqlalchemy import ARRAY, TIMESTAMP, VARCHAR, Boolean, Column, Date, DateTime, Enum, Float, Integer
from sqlalchemy.orm import Session, declarative_base

from chalk import batch
from chalk.features import DataFrame, feature, feature_time, features, has_one
from chalk.features.resolver import OfflineResolver
from chalk.sql import BaseSQLSourceProtocol, ChalkQueryProtocol, PostgreSQLSource, SnowflakeSource, SQLiteFileSource
from chalk.sql._internal.integrations.postgres import PostgreSQLSourceImpl
from chalk.sql._internal.integrations.sqlite import SQLiteFileSourceImpl

try:
    import psycopg2
except ImportError:
    psycopg2 = None

try:
    import snowflake
    import snowflake.sqlalchemy
except ImportError:
    snowflake = None

try:
    import polars as pl
except ImportError:
    pl = None


@features
class SQLExecuteNestedFeatures:
    id: int = feature(primary=True)
    nested_int: int


class EnumFeature(enum.Enum):
    RED = "red"
    BLUE = "blue"


@features
class SQLExecuteFeatures:
    id: int = feature(primary=True)
    ts: datetime = feature_time()
    str_feature: str
    int_feature: int
    bool_feature: bool
    enum_feature: EnumFeature
    date_feature: date
    timestamp_feature: datetime
    float_feature: float
    datetime_feature: datetime
    list_feature: List[str]
    nested_features_id: int
    nested_features: SQLExecuteNestedFeatures = has_one(
        lambda: SQLExecuteFeatures.nested_features_id == SQLExecuteNestedFeatures.id
    )


@pytest.fixture
def sql_source(request, tmp_path: pathlib.Path, monkeypatch: pytest.MonkeyPatch):
    if request.param == "sqlite":
        return SQLiteFileSource(str(tmp_path / "db.sqlite"))
    if request.param in ("postgres"):
        if "PGHOST" not in os.environ:
            pytest.skip("No PGHOST is specified; skipping postgres tests")
        if psycopg2 is None:
            pytest.skip("psycopg2 is not installed")
        return PostgreSQLSource()
    if request.param == "snowflake":
        if "SNOWFLAKE_DATABASE" not in os.environ:
            pytest.skip("No SNOWFLAKE_DATABASE is specified; skipping snowflake tests")
        if snowflake is None:
            pytest.skip("Snowflake is not installed")
        return SnowflakeSource()


@pytest.fixture
def model(sql_source: BaseSQLSourceProtocol):
    randr = str(uuid.uuid4())[10:].replace("-", "_")

    Base: Type = declarative_base()

    class ExecuteSQLModel(Base):
        __tablename__ = f"execute_sql_table_test_{randr}"

        id = Column(Integer, primary_key=True)
        str_feature = Column(VARCHAR)
        int_feature = Column(Integer)
        enum_feature = Column(Enum(EnumFeature, name=f"enum_feature_{randr}"))
        enum_as_string_feature = Column(
            VARCHAR
        )  # The enum feature, but manually stored as a string of the enum's value
        date_feature = Column(Date)
        timestamp_feature = Column(TIMESTAMP)

        bool_feature = Column(Boolean)
        float_feature = Column(Float)
        datetime_feature = Column(DateTime)
        nested_features_id = Column(Integer)
        nested_features_nested_int = Column(Integer)
        if isinstance(sql_source, PostgreSQLSourceImpl):
            list_feature = Column(ARRAY(VARCHAR))

    Base.metadata.create_all(sql_source.get_engine())
    try:
        with Session(sql_source.get_engine()) as session, session.begin():
            session.add(
                ExecuteSQLModel(
                    id=1,
                    # Including a newline and quotes in the value to ensure that escaping works properly
                    str_feature='hello\n,"Ravi"',
                    int_feature=2,
                    bool_feature=True,
                    enum_feature=EnumFeature.RED,
                    date_feature=date(2000, 1, 2),
                    timestamp_feature=datetime(2000, 1, 3, tzinfo=timezone.utc),
                    enum_as_string_feature=EnumFeature.BLUE.value,
                    float_feature=3.14,
                    datetime_feature=datetime(2000, 1, 1),
                    **({"list_feature": ["foo", "bar"]} if isinstance(sql_source, PostgreSQLSourceImpl) else {}),
                    nested_features_id=4,
                    nested_features_nested_int=5,
                )
            )
            session.add(
                ExecuteSQLModel(
                    id=2,
                    str_feature=None,
                    int_feature=None,
                    bool_feature=None,
                    enum_feature=None,
                    date_feature=None,
                    timestamp_feature=None,
                    enum_as_string_feature=None,
                    float_feature=None,
                    datetime_feature=None,
                    **({"list_feature": None} if isinstance(sql_source, PostgreSQLSourceImpl) else {}),
                    nested_features_id=None,
                    nested_features_nested_int=None,
                )
            )
        yield ExecuteSQLModel
    finally:
        Base.metadata.drop_all(sql_source.get_engine())


def get_model_to_primitive_features(model: Any):
    return SQLExecuteFeatures(  # type: ignore -- pyright does not understand the constructor
        id=model.id,
        str_feature=model.str_feature,
        bool_feature=model.bool_feature,
        float_feature=model.float_feature,
        int_feature=model.int_feature,
        date_feature=model.date_feature,
        timestamp_feature=model.timestamp_feature,
        datetime_feature=model.datetime_feature,
        enum_feature=model.enum_as_string_feature,
        nested_features_id=model.nested_features_id,
        nested_features=SQLExecuteNestedFeatures(  # type: ignore -- pyright does not understand the constructor
            id=model.nested_features_id,
            nested_int=model.nested_features_nested_int,
        ),
    )


def get_model_to_list_features(model: Any):
    return SQLExecuteFeatures(  # type: ignore -- pyright does not understand the constructor
        id=model.id,
        list_feature=model.list_feature,
    )


def get_model_to_enum_features(model: Any):
    return SQLExecuteFeatures(  # type: ignore -- pyright does not understand the constructor
        id=model.id,
        enum_feature=model.enum_feature,
    )


@pytest.fixture
def graph_list(sql_source: BaseSQLSourceProtocol, model: Any, request: pytest.FixtureRequest):
    string_query: bool = request.param

    @batch
    def sql_execute_resolve_all_list() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.list_feature,
    ]:
        if string_query:
            return sql_source.query_string(f"select id, list_feature from {model.__tablename__}")
        else:
            return sql_source.query(get_model_to_list_features(model))

    @batch
    def sql_execute_resolve_not_null_list() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.list_feature,
    ]:
        if string_query:
            return sql_source.query_string(f"select id, list_feature from {model.__tablename__} where id = 1")
        else:
            return sql_source.query(get_model_to_list_features(model)).filter(model.id == 1)

    @batch
    def sql_execute_resolve_null_list() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.list_feature,
    ]:
        if string_query:
            return sql_source.query_string(f"select id, list_feature from {model.__tablename__} where id = 2")
        else:
            return sql_source.query(get_model_to_list_features(model)).filter(model.id == 2)

    return sql_execute_resolve_all_list, sql_execute_resolve_not_null_list, sql_execute_resolve_null_list


@pytest.fixture
def graph_enum(sql_source: BaseSQLSourceProtocol, model: Any, request: pytest.FixtureRequest):
    string_query = request.param

    @batch
    def sql_execute_resolve_all_enum() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.enum_feature,
    ]:
        if string_query:
            return sql_source.query_string(f"select id, enum_feature from {model.__tablename__}")
        else:
            return sql_source.query(get_model_to_enum_features(model))

    @batch
    def sql_execute_resolve_not_null_enum() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.enum_feature,
    ]:
        if string_query:
            return sql_source.query_string(f"select id, enum_feature from {model.__tablename__} where id = 1")
        else:
            return sql_source.query(get_model_to_enum_features(model)).filter(model.id == 1)

    @batch
    def sql_execute_resolve_null_enum() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.enum_feature,
    ]:
        if string_query:
            return sql_source.query_string(f"select id, enum_feature from {model.__tablename__} where id = 2")
        else:
            return sql_source.query(get_model_to_enum_features(model)).filter(model.id == 2)

    return sql_execute_resolve_all_enum, sql_execute_resolve_not_null_enum, sql_execute_resolve_null_enum


@pytest.fixture
def graph(sql_source: BaseSQLSourceProtocol, model: Any, request: pytest.FixtureRequest):
    string_query: bool = request.param

    @batch
    def sql_execute_resolve_all() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.str_feature,
        SQLExecuteFeatures.bool_feature,
        SQLExecuteFeatures.float_feature,
        SQLExecuteFeatures.int_feature,
        SQLExecuteFeatures.date_feature,
        SQLExecuteFeatures.timestamp_feature,
        SQLExecuteFeatures.datetime_feature,
        SQLExecuteFeatures.enum_feature,
        SQLExecuteFeatures.nested_features_id,
        SQLExecuteFeatures.nested_features.id,
        SQLExecuteFeatures.nested_features.nested_int,
    ]:
        if string_query:
            return sql_source.query_string(
                f"""select
                id,
                str_feature,
                bool_feature,
                float_feature,
                int_feature,
                date_feature,
                timestamp_feature,
                datetime_feature,
                enum_as_string_feature as enum_feature,
                nested_features_id,
                nested_features_id as "nested_features.id",
                nested_features_nested_int as "nested_features.nested_int"
                from {model.__tablename__}"""
            )
        else:
            return sql_source.query(get_model_to_primitive_features(model))

    @batch
    def sql_execute_resolve_not_null() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.str_feature,
        SQLExecuteFeatures.bool_feature,
        SQLExecuteFeatures.int_feature,
        SQLExecuteFeatures.date_feature,
        SQLExecuteFeatures.timestamp_feature,
        SQLExecuteFeatures.float_feature,
        SQLExecuteFeatures.datetime_feature,
        SQLExecuteFeatures.enum_feature,
        SQLExecuteFeatures.nested_features_id,
        SQLExecuteFeatures.nested_features.id,
        SQLExecuteFeatures.nested_features.nested_int,
    ]:
        if string_query:
            return sql_source.query_string(
                f"""select
                id,
                str_feature,
                bool_feature,
                float_feature,
                int_feature,
                date_feature,
                timestamp_feature,
                datetime_feature,
                enum_as_string_feature as enum_feature,
                nested_features_id,
                nested_features_id as "nested_features.id",
                nested_features_nested_int as "nested_features.nested_int"
                from {model.__tablename__} where id = 1"""
            )
        else:
            return sql_source.query(get_model_to_primitive_features(model)).filter(model.id == 1)

    @batch
    def sql_execute_resolve_null() -> DataFrame[
        SQLExecuteFeatures.id,
        SQLExecuteFeatures.str_feature,
        SQLExecuteFeatures.bool_feature,
        SQLExecuteFeatures.int_feature,
        SQLExecuteFeatures.date_feature,
        SQLExecuteFeatures.timestamp_feature,
        SQLExecuteFeatures.enum_feature,
        SQLExecuteFeatures.float_feature,
        SQLExecuteFeatures.datetime_feature,
        SQLExecuteFeatures.nested_features_id,
        SQLExecuteFeatures.nested_features.id,
        SQLExecuteFeatures.nested_features.nested_int,
    ]:
        if string_query:
            return sql_source.query_string(
                f"""select
                id,
                str_feature,
                bool_feature,
                float_feature,
                int_feature,
                date_feature,
                timestamp_feature,
                datetime_feature,
                enum_as_string_feature as enum_feature,
                nested_features_id,
                nested_features_id as "nested_features.id",
                nested_features_nested_int as "nested_features.nested_int"
                from {model.__tablename__} where id = 2"""
            )
        else:
            return sql_source.query(get_model_to_primitive_features(model)).filter(model.id == 2)

    return sql_execute_resolve_all, sql_execute_resolve_not_null, sql_execute_resolve_null


@pytest.fixture
def force_sqlalchemy_query_execution(request: pytest.FixtureRequest, monkeypatch: pytest.MonkeyPatch) -> bool:
    if request.param:
        monkeypatch.setenv("CHALK_FORCE_SQLALCHEMY_QUERY_EXECUTION", "1")
    return request.param


@pytest.mark.parametrize("sql_source", ["postgres", "sqlite", "snowflake"], indirect=True)
@pytest.mark.parametrize("force_sqlalchemy_query_execution", [True, False], indirect=True)
@pytest.mark.skipif(pl is None, reason="polars is not installed")
class TestExecute:
    def _test_harness(
        self,
        resolver: OfflineResolver[Any, DataFrame],
        expected: DataFrame,
    ):
        # Test that calling the resolver directly returns a dataframe
        direct_query = resolver()
        assert isinstance(direct_query, DataFrame)
        assert (direct_query[expected.columns] == expected).all()

        # Also test manually invoking via execute_to_datafrmae
        indirect_query = cast(ChalkQueryProtocol, resolver.fn())
        finalized_query = indirect_query.all()
        if isinstance(resolver.output.features[0], type) and issubclass(resolver.output.features[0], DataFrame):
            features = resolver.output.features[0].columns
        else:
            features = resolver.output.features
        output_df = finalized_query.execute_to_dataframe(features)
        assert (output_df == expected).all()

    def _assert_efficient_execution(self, caplog: pytest.LogCaptureFixture):
        for logger_name, log_level, message in caplog.record_tuples:
            assert "The SQL statement will be executed into SQLAlchemy objects" not in message

    def _assert_sqlalchemy_execution(self, caplog: pytest.LogCaptureFixture):
        found = False
        for logger_name, log_level, message in caplog.record_tuples:
            if "The SQL statement will be executed into SQLAlchemy objects" in message:
                found = True
        assert found

    @pytest.mark.parametrize("graph", [True, False], indirect=True)
    def test_resolve_all(
        self,
        graph: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        caplog: pytest.LogCaptureFixture,
        force_sqlalchemy_query_execution: bool,
        sql_source: BaseSQLSourceProtocol,
    ):
        caplog.set_level(logging.INFO)

        sql_execute_resolve_all = graph[0]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [1, 2],
                SQLExecuteFeatures.str_feature: ['hello\n,"Ravi"', None],
                SQLExecuteFeatures.int_feature: [2, None],
                SQLExecuteFeatures.bool_feature: [True, None],
                SQLExecuteFeatures.enum_feature: [EnumFeature.BLUE, None],
                SQLExecuteFeatures.date_feature: [datetime(2000, 1, 2), None],
                SQLExecuteFeatures.timestamp_feature: [datetime(2000, 1, 3, tzinfo=timezone.utc), None],
                SQLExecuteFeatures.float_feature: [3.14, None],
                SQLExecuteFeatures.datetime_feature: [datetime(2000, 1, 1), None],
                SQLExecuteFeatures.nested_features_id: [4, None],
                SQLExecuteFeatures.nested_features.id: [4, None],
                SQLExecuteFeatures.nested_features.nested_int: [5, None],
            }
        )
        self._test_harness(sql_execute_resolve_all, expected)
        if force_sqlalchemy_query_execution or isinstance(sql_source, SQLiteFileSourceImpl):
            self._assert_sqlalchemy_execution(caplog)
        else:
            self._assert_efficient_execution(caplog=caplog)

    @pytest.mark.parametrize("graph", [True, False], indirect=True)
    def test_resolve_not_null(
        self,
        graph: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        caplog: pytest.LogCaptureFixture,
        force_sqlalchemy_query_execution: bool,
        sql_source: BaseSQLSourceProtocol,
    ):
        caplog.set_level(logging.INFO)
        sql_execute_resolve_not_null = graph[1]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [1],
                SQLExecuteFeatures.str_feature: ['hello\n,"Ravi"'],
                SQLExecuteFeatures.int_feature: [2],
                SQLExecuteFeatures.bool_feature: [True],
                SQLExecuteFeatures.enum_feature: [EnumFeature.BLUE],
                SQLExecuteFeatures.float_feature: [3.14],
                SQLExecuteFeatures.date_feature: [datetime(2000, 1, 2)],
                SQLExecuteFeatures.timestamp_feature: [datetime(2000, 1, 3, tzinfo=timezone.utc)],
                SQLExecuteFeatures.datetime_feature: [datetime(2000, 1, 1)],
                SQLExecuteFeatures.nested_features_id: [4],
                SQLExecuteFeatures.nested_features.id: [4],
                SQLExecuteFeatures.nested_features.nested_int: [5],
            }
        )
        self._test_harness(sql_execute_resolve_not_null, expected)
        if force_sqlalchemy_query_execution or isinstance(sql_source, SQLiteFileSourceImpl):
            self._assert_sqlalchemy_execution(caplog)
        else:
            self._assert_efficient_execution(caplog=caplog)

    @pytest.mark.parametrize("graph", [True, False], indirect=True)
    def test_resolve_null(
        self,
        graph: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        caplog: pytest.LogCaptureFixture,
        force_sqlalchemy_query_execution: bool,
        sql_source: BaseSQLSourceProtocol,
    ):
        caplog.set_level(logging.INFO)
        sql_execute_resolve_null = graph[2]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [2],
                SQLExecuteFeatures.str_feature: [None],
                SQLExecuteFeatures.int_feature: [None],
                SQLExecuteFeatures.bool_feature: [None],
                SQLExecuteFeatures.enum_feature: [None],
                SQLExecuteFeatures.date_feature: [None],
                SQLExecuteFeatures.timestamp_feature: [None],
                SQLExecuteFeatures.float_feature: [None],
                SQLExecuteFeatures.datetime_feature: [None],
                SQLExecuteFeatures.nested_features_id: [None],
                SQLExecuteFeatures.nested_features.id: [None],
                SQLExecuteFeatures.nested_features.nested_int: [None],
            }
        )
        self._test_harness(sql_execute_resolve_null, expected)
        if force_sqlalchemy_query_execution or isinstance(sql_source, SQLiteFileSourceImpl):
            self._assert_sqlalchemy_execution(caplog)
        else:
            self._assert_efficient_execution(caplog=caplog)

    @pytest.mark.parametrize(
        "graph_list",
        [pytest.param(True, marks=pytest.mark.xfail(reason="String list queries don't work")), False],
        indirect=True,
    )
    def test_resolve_all_list(
        self,
        graph_list: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        sql_source: BaseSQLSourceProtocol,
        force_sqlalchemy_query_execution: bool,
    ):
        if not isinstance(sql_source, PostgreSQLSourceImpl):
            pytest.skip("Only postgres supports arrays")
        sql_execute_resolve_all = graph_list[0]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [1, 2],
                SQLExecuteFeatures.list_feature: [["foo", "bar"], None],
            }
        )
        self._test_harness(sql_execute_resolve_all, expected)

    @pytest.mark.parametrize(
        "graph_list",
        [pytest.param(True, marks=pytest.mark.xfail(reason="String list queries don't work")), False],
        indirect=True,
    )
    def test_resolve_not_null_list(
        self,
        graph_list: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        sql_source: BaseSQLSourceProtocol,
        force_sqlalchemy_query_execution: bool,
    ):
        if not isinstance(sql_source, PostgreSQLSourceImpl):
            pytest.skip("Only postgres supports arrays")
        sql_execute_resolve_not_null = graph_list[1]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [1],
                SQLExecuteFeatures.list_feature: [["foo", "bar"]],
            }
        )
        self._test_harness(sql_execute_resolve_not_null, expected)

    @pytest.mark.parametrize("graph_list", [True, False], indirect=True)
    def test_resolve_null_list(
        self,
        graph_list: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        sql_source: BaseSQLSourceProtocol,
        force_sqlalchemy_query_execution: bool,
    ):
        if not isinstance(sql_source, PostgreSQLSourceImpl):
            pytest.skip("Only postgres supports arrays")
        sql_execute_resolve_null = graph_list[2]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [2],
                SQLExecuteFeatures.list_feature: [None],
            }
        )
        self._test_harness(sql_execute_resolve_null, expected)

    @pytest.mark.parametrize(
        "graph_enum",
        [pytest.param(True, marks=pytest.mark.xfail(reason="Enum string queries don't work")), False],
        indirect=True,
    )
    def test_resolve_all_enum(
        self,
        graph_enum: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        force_sqlalchemy_query_execution: bool,
    ):
        sql_execute_resolve_all = graph_enum[0]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [1, 2],
                SQLExecuteFeatures.enum_feature: [EnumFeature.RED, None],
            }
        )
        self._test_harness(sql_execute_resolve_all, expected)

    @pytest.mark.parametrize(
        "graph_enum",
        [pytest.param(True, marks=pytest.mark.xfail(reason="Enum string queries don't work")), False],
        indirect=True,
    )
    def test_resolve_not_null_enum(
        self,
        graph_enum: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        force_sqlalchemy_query_execution: bool,
    ):
        sql_execute_resolve_not_null = graph_enum[1]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [1],
                SQLExecuteFeatures.enum_feature: [EnumFeature.RED],
            }
        )
        self._test_harness(sql_execute_resolve_not_null, expected)

    @pytest.mark.parametrize("graph_enum", [True, False], indirect=True)
    def test_resolve_null_enum(
        self,
        graph_enum: Tuple[OfflineResolver, OfflineResolver, OfflineResolver],
        force_sqlalchemy_query_execution: bool,
    ):
        sql_execute_resolve_null = graph_enum[2]
        expected = DataFrame(
            {
                SQLExecuteFeatures.id: [2],
                SQLExecuteFeatures.enum_feature: [None],
            }
        )
        self._test_harness(sql_execute_resolve_null, expected)



from typing import List

from chalk.client.models import ChalkError


class ChalkBaseException(Exception):
    ...


class ChalkOfflineQueryException(ChalkBaseException):
    message: str
    errors: List[ChalkError]

    def __init__(self, message: str, errors: List[ChalkError]):
        self.message = message
        self.errors = errors
        super().__init__(message + "\n" + "\n".join(["\t" + e.message for e in errors[0:3]]))


class ChalkResolverRunException(ChalkBaseException):
    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


class ChalkDatasetDownloadException(ChalkBaseException):
    message: str

    def __init__(self, message: str):
        self.message = message
        super().__init__(message)


__all__ = [
    "ChalkBaseException",
    "ChalkOfflineQueryException",
    "ChalkResolverRunException",
    "ChalkDatasetDownloadException",
]



import re
from typing import List


def to_snake_case(name: str):
    name = re.sub("(.)([A-Z][a-z]+)", r"\1_\2", name)
    name = re.sub("__([A-Z])", r"_\1", name)
    name = re.sub("([a-z0-9])([A-Z])", r"\1_\2", name)
    return name.lower()


def removeprefix(text: str, prefix: str):
    """
    This functionality is available in Python 3.9 as text.removeprefix(prefix),
    but we're supporting down to 3.8, where it is not available.
    """
    return text[len(prefix) :] if text.startswith(prefix) else text


def comma_whitespace_split(value: str):
    return re.split(r"\s*,\s*", value)


def normalize_string_for_matching(name: str) -> str:
    """
    1. Case-insensitive
    2. Strip everything except alphanumeric or periods, so e.g. `first_name` matches `firstname`
    """
    return re.sub(r"[^0-9a-z\.]", "", name.lower())


def to_camel_case(snake_str: str):
    components = [c for c in snake_str.split("_") if c != ""]
    return components[0].lower() + "".join(x.title() for x in components[1:])


def to_title_case(snake_str: str):
    cameled = to_camel_case(snake_str)
    return cameled[0].upper() + cameled[1:] if len(cameled) > 0 else cameled


def oxford_comma_list(lst: List[str]) -> str:
    if len(lst) == 0:
        return ""
    elif len(lst) == 1:
        return lst[0]
    elif len(lst) == 2:
        return f"{lst[0]} and {lst[1]}"

    return ", ".join(lst[:-1]) + f", and {lst[-1]}"



from chalk.features import Primary, feature, features, is_primary


@features
class PrimaryCls:
    myid: Primary[str]


@features
class PrimaryCls2:
    myid: Primary[str] = feature(max_staleness="3d")
    id: str


def test_primary():
    assert is_primary(PrimaryCls.myid)
    assert is_primary(PrimaryCls2.myid)
    assert not is_primary(PrimaryCls2.id)



from datetime import datetime

from chalk.features import (
    features,
    FeatureTime,
    is_feature_time,
    feature_time,
)


@features
class FTCls:
    id: str
    timestamp: FeatureTime


def test_by_annotation():
    assert is_feature_time(FTCls.timestamp)
    assert not is_feature_time(FTCls.id)


@features
class FTCls1:
    id: str
    ts: datetime


def test_by_naming_convention():
    assert is_feature_time(FTCls1.ts)


@features
class FTCls2:
    id: str
    ts: datetime
    timestamp: datetime = feature_time()


def test_by_feature():
    assert is_feature_time(FTCls2.timestamp)
    assert not is_feature_time(FTCls2.ts)



from datetime import timedelta

from chalk.utils.duration import timedelta_to_duration


def test_timedelta_to_duration():
    assert timedelta_to_duration(timedelta(seconds=4)) == "4s"
    assert timedelta_to_duration(timedelta(seconds=4, minutes=30)) == f"{30*60+4}s"
    assert timedelta_to_duration(timedelta(seconds=4, microseconds=44)) == f"4s"
    assert timedelta_to_duration(timedelta(seconds=4, microseconds=44, days=1)) == f"1d 4s"
    assert timedelta_to_duration(timedelta(days=2)) == f"2d"
